searchData={"items":[{"type":"module","title":"ProcessHub","doc":"This is the main public API module for the `ProcessHub` library.","ref":"ProcessHub.html"},{"type":"module","title":"Table of Contents - ProcessHub","doc":"- [Description](#module-description)\n- [Features](#module-features)\n- [Installation](#module-installation)\n- [Example usage](#module-example-usage)\n- [Configurable strategies](#module-configuration)\n- [Distribution strategy](#module-default-distribution-strategy)\n- [Cluster discovery and formation](#module-cluster-discovery-and-formation)\n- [Resilience and reliability](#module-resilience-and-reliability)\n- [Locking mechanism](#module-locking-mechanism)\n- [Hooks](#module-hooks)\n- [Contributing](#module-contributing)\n- [Types](#types)\n- [Public API functions](#functions)","ref":"ProcessHub.html#module-table-of-contents"},{"type":"module","title":"Description - ProcessHub","doc":"Library for building distributed systems that are scalable. It handles the distribution of\nprocesses within a cluster of nodes while providing a globally synchronized process registry.\n\nProcessHub takes care of starting, stopping and monitoring processes in the cluster.\nIt scales automatically when cluster is updated and handles network partitions.\n\nBuilding distributed systems is hard and designing one is all about trade-offs.\nThere are many things to consider and each system has its own requirements.\nThis library aims to be flexible and configurable to suit different use cases.\n\nProcessHub is designed to be **decentralized** in its architecture. It does not rely on a\nsingle node to manage the cluster. Each node in the cluster is considered equal.\nConsensus is achieved by using a hash ring implementation.\n\n> ProcessHub is built with scalability and availability in mind.\n> Most of the operations are asynchronous and non-blocking.\n> It can guarantee **eventual** consistency.\n\nProcessHub provides a set of configurable strategies for building distributed\napplications in Elixir.\n\n> ##","ref":"ProcessHub.html#module-description"},{"type":"module","title":"ProcessHub requires a distributed node {: .info} - ProcessHub","doc":"> ProcessHub is distributed in its nature, and for that reason, it needs to\n> **operate in a distributed environment**.\n> This means that the Elixir instance has to be started as a distributed node.\n> For example: `iex --sname mynode --cookie mycookie -S mix`.\n>\n> If the node is not started as a distributed node, starting the `ProcessHub` will fail\n> with the following error: `{:error, :local_node_not_alive}`","ref":"ProcessHub.html#module-processhub-requires-a-distributed-node-info"},{"type":"module","title":"Features - ProcessHub","doc":"Main features include:\n- Distribute processes within a cluster of nodes.\n- Provides globally synchronized process registry.\n- Automatic hub cluster forming and healing when nodes join or leave the cluster.\n- Process state handover.\n- Strategies for redundancy handling and process replication.\n- Strategies for handling network failures and partitions automatically.\n- Strategies for handling process migration and synchronization when nodes join/leave\nthe cluster automatically.\n- Hooks for triggering events on specific actions.","ref":"ProcessHub.html#module-features"},{"type":"module","title":"Installation - ProcessHub","doc":"1. Add `process_hub` to your list of dependencies in `mix.exs`:\n\n    ```elixir\n    def deps do\n      [\n        {:process_hub, \"~> 0.1.3-alpha\"}\n      ]\n    end\n    ```\n\n2. Start the `ProcessHub` supervisor under your application supervision tree:\n\n    ```elixir\n    defmodule MyApp.Application do\n      use Application\n\n      def start(_type, _args) do\n        children = [\n          ProcessHub.child_spec(%ProcessHub{hub_id: :my_hub})\n        ]\n\n        opts = [strategy: :one_for_one, name: MyApp.Supervisor]\n        Supervisor.start_link(children, opts)\n      end\n    end\n    ```\n  It is possible to start multiple hubs under the same supervision tree.\n  Each hub must have a unique `t:hub_id/0`.","ref":"ProcessHub.html#module-installation"},{"type":"module","title":"Example usage - ProcessHub","doc":"The following example shows how to start 2 elixir nodes, connect them and start processes\nunder the `ProcessHub` cluster. This demonstrates how the processes are distributed within\nthe cluster.\n\n**Note:** The examples below assume that the `ProcessHub` is already started under the\nsupervision tree. If not please refer to the [Installation](#module-installation) section.\n\n**Note:** Make sure you have a GenServer module called `MyProcess` defined in your project.\n```elixir\ndefmodule MyProcess do\n  use GenServer\n\n  def start_link() do\n    GenServer.start_link(__MODULE__, nil)\n  end\n\n  def init(_) do\n    {:ok, nil}\n  end\nend\n```\n\n \n\n#","ref":"ProcessHub.html#module-example-usage"},{"type":"module","title":"Node 1 - ProcessHub","doc":"Start the first node with the following command:\n\n```bash\niex --name node1@127.0.0.1 --cookie mycookie -S mix\n```\n\n```elixir\n# Run the following in the iex console to start 5 processes under the hub.\niex> ProcessHub.start_children(:my_hub, [\n...>  %{id: :some_id1, start: {MyProcess, :start_link, []}},\n...>  %{id: :another_id2, start: {MyProcess, :start_link, []}},\n...>  %{id: :child_3, start: {MyProcess, :start_link, []}},\n...>  %{id: :child_4, start: {MyProcess, :start_link, []}},\n...>  %{id: \"the_last_child\", start: {MyProcess, :start_link, []}}\n...> ])\n{:ok, :start_initiated}\n\n# Check the started processes by running the command below.\niex> ProcessHub.which_children(:my_hub, [:global])\n[\n  node1@127.0.0.1: [\n    {\"the_last_child\", #PID , :worker, [MyProcess]},\n    {:child_4, #PID , :worker, [MyProcess]},\n    {:child_3, #PID , :worker, [MyProcess]},\n    {:another_id2, #PID , :worker, [MyProcess]},\n    {:some_id1, #PID , :worker, [MyProcess]}\n  ]\n]\n```\n\n#","ref":"ProcessHub.html#module-node-1"},{"type":"module","title":"Node 2 - ProcessHub","doc":"We will use this node to connect to the first node and see how the processes are\nautomatically distributed.\n\nStart the second node.\n```bash\niex --name node2@127.0.0.1 --cookie mycookie -S mix\n```\n\n```elixir\n# Connect the second node to the first node.\niex> Node.connect(:\"node1@127.0.0.1\")\ntrue\n\n# Check the started procsses by running the command below and\n# see how some of the processes are distributed to the second node.\niex> ProcessHub.which_children(:my_hub, [:global])\n[\n  \"node2@127.0.0.1\": [\n    {:child_3, #PID , :worker, [MyProcess]},\n    {:some_id1, #PID , :worker, [MyProcess]}\n  ],\n  \"node1@127.0.0.1\": [\n    {\"the_last_child\", #PID , :worker, [MyProcess]},\n    {:child_4, #PID , :worker, [MyProcess]},\n    {:another_id2, #PID , :worker, [MyProcess]}\n  ]\n]\n```","ref":"ProcessHub.html#module-node-2"},{"type":"module","title":"Configurable strategies - ProcessHub","doc":"ProcessHub comes with 9 different strategies that can be used to configure the hub.\nAll strategies are Elixir structs that implement their own base protocol.\n\nIn fact, you can define your own strategies by implementing the base protocols.\n\nWhen configuring the hub, you can pass the strategies as part of the `t:t/0` struct.\n\nLook at the documentation for each strategy for more information on how to configure them.\n\nAn example can be seen below.\n```elixir\ndefmodule MyApp.Application do\n  use Application\n\n  def start(_type, _args) do\n    children = [process_hub()]\n\n    opts = [strategy: :one_for_one, name: MyApp.Supervisor]\n    Supervisor.start_link(children, opts)\n  end\n\n  defp process_hub() do\n    {ProcessHub, %ProcessHub{\n      hub_id: :my_hub,\n      # Configure the redundancy strategy.\n      redundancy_strategy: %ProcessHub.Strategy.Redundancy.Replication{\n        replication_factor: 2,\n        replication_model: :active_passive,\n        redundancy_signal: :all\n      },\n      # Configure the migration strategy.\n      migration_strategy: %ProcessHub.Strategy.Migration.HotSwap{\n        retention: 2000,\n        handover: true\n      },\n      # Configure the synchronization strategy.\n      synchronization_strategy: %ProcessHub.Strategy.Synchronization.PubSub{\n        sync_interval: 10000\n      },\n      # Configure the partition tolerance strategy.\n      partition_tolerance_strategy: %ProcessHub.Strategy.PartitionTolerance.StaticQuorum{\n        quorum_size: 2\n      }\n    }}\n  end\nend\n```\n\n#","ref":"ProcessHub.html#module-configurable-strategies"},{"type":"module","title":"Redundancy Strategy - ProcessHub","doc":"`ProcessHub.Strategy.Redundancy.Base` - defines the base protocol for redundancy strategies.\nThis strategy is used to define how many replicas of a process should be started\nacross the cluster. Starting multiple instances of a process across the cluster is useful\nfor redundancy and fault tolerance.\n\nAvailable strategies are:\n- `ProcessHub.Strategy.Redundancy.Singularity` - only 1 process per `t:child_id/0` without\nany replicas. This is also the default strategy and contains no special configuration options.\n- `ProcessHub.Strategy.Redundancy.Replication` - starts multiple replicas of a process\nacross the cluster. The number of replicas is defined by the `:replication_factor`\noption. This strategy also supports `:active_active` and `:active_passive` modes.\nMeaning we may have one active process and the rest are passive.\nThe mode is defined by the `:replication_model` option.\nThis information will be passed to the started process.\nThe default mode is `:active_active`, meaning all processes are equal and considered active.\n\n#","ref":"ProcessHub.html#module-redundancy-strategy"},{"type":"module","title":"Migration Strategy - ProcessHub","doc":"`ProcessHub.Strategy.Migration.Base` - defines the base protocol for migration strategies.\nThis strategy is used to define how the processes are migrated when a node joins or leaves the cluster.\n\nMigration is the process of moving processes from one node to another.\nOne of the reasons why migration happens is when a node leaves the cluster.\nWhen a node leaves the cluster, it is possible that some processes are still running on that\nnode, so these need to be migrated to another node. Also, when a new node joins the\ncluster, other nodes may migrate some processes over to the new node.\n\nAt the moment, there are 2 migration strategies available:\n- `ProcessHub.Strategy.Migration.ColdSwap` - migrate processes by stopping the process\non the old node before starting it on the new node. This is the default strategy and defines no special configuration options.\n- `ProcessHub.Strategy.Migration.HotSwap` - this strategy is used to migrate processes\nby starting the process on the new node before stopping it on the old node.\nThis strategy is useful when we want to avoid any downtime. This strategy is also\nuseful when the process is stateful, and we want to avoid any data loss by handing over\nthe state from the old process to the new process. See the module for handover examples.\n\n#","ref":"ProcessHub.html#module-migration-strategy"},{"type":"module","title":"Synchronization Strategy - ProcessHub","doc":"`ProcessHub.Strategy.Synchronization.Base` - defines the base protocol for synchronization\nstrategies which define the method that is used to synchronize the process registry.\n\nAvailable strategies are:\n- `ProcessHub.Strategy.Synchronization.PubSub` - uses a publish/subscribe model to synchronize\nthe process registry. Each node in the cluster will subscribe to a topic and publish any changes to the topic.\nThese changes could be events such as adding or removing a process from the registry.\nThis is the default and recommended synchronization strategy for most users.\n- `ProcessHub.Strategy.Synchronization.Gossip` - uses a gossip protocol to synchronize the\nprocess registry. Using this strategy is recommended when the number of nodes in the cluster\nis large. The Gossip strategy selects a predefined number of nodes to gossip with and\nexchange information about the process registry.\nThese selected nodes will choose other nodes to gossip with and so on until all nodes in the\ncluster are synchronized. It has higher latency than the PubSub strategy specially when the cluster\nis rather small.\n\n#","ref":"ProcessHub.html#module-synchronization-strategy"},{"type":"module","title":"Partition Tolerance Strategy - ProcessHub","doc":"`ProcessHub.Strategy.PartitionTolerance.Base` - defines the base protocol for partition tolerance\nstrategies which define the method that is used to handle network partitions.\n\nAvailable strategies are:\n- `ProcessHub.Strategy.PartitionTolerance.Divergence` - this strategy is used to handle network\npartitions by diverging the cluster into multiple subclusters. Each subcluster will have its\nown hub and will be considered as a separate cluster. This strategy is the default strategy.\nWhen the network partition is healed, the subclusters will merge back into a single cluster.\n- `ProcessHub.Strategy.PartitionTolerance.StaticQuorum` - this strategy is used to handle network\npartitions by using a static quorum. The quorum size is defined by the `:quorum_size` option.\nWhen a partition happens, the `ProcessHub.DistributedSupervisor` process will terminate\nalong with its children. This strategy is useful when the number of nodes in the cluster is\nknown and rather fixed.\n- `ProcessHub.Strategy.PartitionTolerance.DynamicQuorum` - this strategy is used to handle\nnetwork partitions by using a dynamic quorum. The quorum size is defined by the `:quorum_size`\noption and `:threshold_time` option. The system automatically over time adapts to the number of\nnodes in the cluster. When a partition happens, the `ProcessHub.DistributedSupervisor` process\nwill terminate along with its children.\n  > ##","ref":"ProcessHub.html#module-partition-tolerance-strategy"},{"type":"module","title":"Using DynamicQuorum Strategy {: .error} - ProcessHub","doc":"> When scaling down too many nodes at once, the system may consider itself to be in a\n  > network partition. Read the documentation for the\n  >`ProcessHub.Strategy.PartitionTolerance.DynamicQuorum` strategy for more information.","ref":"ProcessHub.html#module-using-dynamicquorum-strategy-error"},{"type":"module","title":"Distribution Strategy - ProcessHub","doc":"ProcessHub uses consistent hashing to distribute processes. When the cluster is updated, the\nhash ring is recalculated. The recalculation is done in a way that each node is assigned a unique\nhash value, and they form a **hash ring**. Each node in the cluster keeps track of the ProcessHub\ncluster and updates its local hash ring accordingly.\n\nTo find the node that the process belongs to, the system will use the hash ring to calculate\nthe hash value of the process ID (`t:child_id/0`) and assign it to the node with the closest hash value.\n\nWhen the cluster is updated and the hash ring is recalculated, it does not mean that all\nprocesses will be shuffled. Only the processes that are affected by the change will be\nredistributed. This is done to avoid unnecessary process migrations.\n\nFor example, when a node leaves the cluster, only the processes that were running on that node\nwill be redistributed. The rest of the processes will stay on the same node. When a new node\njoins the cluster, only some of the processes will be redistributed to the new node, and the\nrest will stay on the same node.\n\n> The hash ring implementation **does not guarantee** that all processes will always be\n> evenly distributed, but it does its best to distribute them as evenly as possible.\n\nThis strategy is used by default and is not configurable at the moment.","ref":"ProcessHub.html#module-distribution-strategy"},{"type":"module","title":"Cluster Discovery and Formation - ProcessHub","doc":"ProcessHub monitors connecting and disconnecting nodes and forms a cluster automatically\nfrom the connected nodes that share the same `t:hub_id/0`. It's not required to start\nthe `ProcessHub` on all nodes in the cluster.","ref":"ProcessHub.html#module-cluster-discovery-and-formation"},{"type":"module","title":"Resilience and Reliability - ProcessHub","doc":"ProcessHub uses the `Supervisor` behavior and leverages the features that come with it.\nEach hub starts its own `ProcessHub.DistributedSupervisor` process, which is responsible for\nstarting, stopping, and monitoring the processes in its local cluster.\n\nWhen a process dies unexpectedly, the `ProcessHub.DistributedSupervisor` will restart it\nautomatically.\n\nProcessHub also takes care of validating the `t:child_spec/0` before starting it and makes sure\nit's started on the right node that the process belongs to.\nIf the process is being started on the wrong node, the initialization request will be forwarded\nto the correct node.","ref":"ProcessHub.html#module-resilience-and-reliability"},{"type":"module","title":"Locking Mechanism - ProcessHub","doc":"ProcessHub utilizes the `:blockade` library to provide event-driven communication\nand a locking mechanism.\nIt locks the local event queue by increasing its priority for some operations.\nThis allows the system to queue events and process them in order to preserve data integrity.\nOther events can be processed once the priority level is set back to default.\n\nTo avoid deadlocks, the system places a timeout on the event queue priority and\nrestores it to its original value if the timeout is reached.","ref":"ProcessHub.html#module-locking-mechanism"},{"type":"module","title":"Hooks - ProcessHub","doc":"Hooks are used to trigger events on specific actions. Hooks can be registered by passing the\nhandlers to the `:hooks` option of the `t:t/0` configuration struct or by inserting them\ndynamically using the `ProcessHub.Service.HookManager` module.\n\nProcessHub heavily uses hooks internally in the integration tests.\n\nHooks have to be in the format of an `mfa` tuple. Basically, they are functions that will be called\nwhen the hook is triggered.\n\nIt is possible to register a hook handler with a wildcard argument `:_`, which will be replaced\nwith the hook data when the hook is dispatched.\n\nExample hook registration using the `t:t/0` configuration struct:\n```elixir\n# Register a hook handler for the `:cluster_join` event with a wildcard argument.\ndefmodule MyApp.Application do\n  use Application\n\n  def start(_type, _args) do\n    children = [\n      ProcessHub.child_spec(%ProcessHub{\n        hub_id: :my_hub,\n        hooks: [\n           ProcessHub.Constant.Hook.cluster_join(), {MyModule, :my_function, [:some_data, :_]}\n        ]\n      })\n    ]\n\n    opts = [strategy: :one_for_one, name: MyApp.Supervisor]\n    Supervisor.start_link(children, opts)\n  end\nend\n```\n\nExample hook registration using the `ProcessHub.Service.HookManager` module:\n```elixir\n# Register a hook handler for the `:cluster_join` event with a wildcard argument.\nProcessHub.Service.HookManager.register_hook(:my_hub, ProcessHub.Constant.Hook.cluster_join(), {MyModule, :my_function, [:some_data, :_]})\n```\n\nExample hook handler:\n```elixir\n# The hook handler should be in the following format:\ndefmodule MyModule do\n  def my_function(:some_data, dynamic_hook_data) do\n    # Do something with the data.\n  end\nend\n```\n\n#","ref":"ProcessHub.html#module-hooks"},{"type":"module","title":"Available hooks - ProcessHub","doc":"| Event Key                   | Trigger                    | Data                                |\n| ------------                | -------------              | ---------------                     |\n| `cluster_join`              | Node joins the cluster     | `node()`                            |\n| `cluster_leave`             | Node leaves the cluster    | `node()`                            |\n| `registry_pid_inserted`     | Process registered         | `{child_id(), [{node(), pid()}]}` |\n| `registry_pid_removed`      | Process unregistered       | `child_id()`                        |\n| `child_migrated`            | Process migrated           | `{child_id(), node()}`              |\n| `priority_state_updated`    | Priority state updated     | `{priority_level(), list()}`        |\n| `pre_nodes_redistribution`  | Nodes redistribution start | `{:nodeup | :nodedown, node()}`     |\n| `post_nodes_redistribution` | Nodes redistribution end   | `{:nodeup |:nodedown, node()}`      |\n\nSee `ProcessHub.Constant.Hook` module for more information.","ref":"ProcessHub.html#module-available-hooks"},{"type":"module","title":"Contributing - ProcessHub","doc":"Contributions are welcome and appreciated. If you have any ideas, suggestions, or bugs to report,\nplease open an issue or a pull request on GitHub.","ref":"ProcessHub.html#module-contributing"},{"type":"function","title":"ProcessHub.await/1","doc":"This function can be used to wait for the `ProcessHub` child start or stop\nfunctions to complete.\n\nThe `await/1` function should be used with the `async_wait: true` option.\n\nKeep in mind that the `await/1` function will block the calling process until\nthe response is received. If the response is not received within the timeout\nperiod, the function will return `{:error, term()}`.","ref":"ProcessHub.html#await/1"},{"type":"function","title":"Example - ProcessHub.await/1","doc":"iex> ref = ProcessHub.start_child(:my_hub, child_spec, [async_wait: true])\n    iex> ProcessHub.await(ref)\n    {:ok, {:my_child, [{:mynode, #PID }]}}","ref":"ProcessHub.html#await/1-example"},{"type":"function","title":"ProcessHub.child_lookup/2","doc":"Returns information about processes that are registered with the given `t:child_id/0`.\n\nThis function queries results from the local `ets` table and does not make any network calls.\n\nThe return results contain the `t:child_spec/0` and a list of tuples where the first element is the node\nwhere the child is started, and the second element is the `pid()` of the started child.","ref":"ProcessHub.html#child_lookup/2"},{"type":"function","title":"Example - ProcessHub.child_lookup/2","doc":"iex> {} = {_child_spec, _node_pid_tuples} = ProcessHub.child_lookup(:my_hub, :my_child)\n    {%{id: :my_child, start: {MyProcess, :start_link, []}}, [{:mynode, #PID }]}","ref":"ProcessHub.html#child_lookup/2-example"},{"type":"function","title":"ProcessHub.child_spec/1","doc":"Returns the child specification for the `ProcessHub.Initializer` supervisor.","ref":"ProcessHub.html#child_spec/1"},{"type":"function","title":"ProcessHub.is_alive?/1","doc":"Checks if the `ProcessHub` with the given `t:hub_id/0` is alive.\n\nA hub is considered alive if the `ProcessHub.Initializer` supervisor process\nis running along with the required child processes for the hub to function.","ref":"ProcessHub.html#is_alive?/1"},{"type":"function","title":"Example - ProcessHub.is_alive?/1","doc":"iex> ProcessHub.is_alive?(:not_existing)\n    false","ref":"ProcessHub.html#is_alive?/1-example"},{"type":"function","title":"ProcessHub.is_locked?/1","doc":"Checks if the `ProcessHub` with the given `t:hub_id/0` is locked.\n\nA hub is considered locked if the `ProcessHub` local event queue has a priority level\ngreater than or equal to 10. This is used to throttle the hub from processing\nany new events and conserve data integrity.","ref":"ProcessHub.html#is_locked?/1"},{"type":"function","title":"Example - ProcessHub.is_locked?/1","doc":"iex> ProcessHub.is_locked?(:my_hub)\n    false","ref":"ProcessHub.html#is_locked?/1-example"},{"type":"function","title":"ProcessHub.is_partitioned?/1","doc":"Checks if the `ProcessHub` with the given `t:hub_id/0` is in a network-partitioned state.\n\nA hub is considered partitioned if the `ProcessHub.Strategy.PartitionTolerance` strategy\nhas detected a network partition. When a network partition is detected, the hub will\nterminate the `ProcessHub.DistributedSupervisor` process along with its children.","ref":"ProcessHub.html#is_partitioned?/1"},{"type":"function","title":"Example - ProcessHub.is_partitioned?/1","doc":"iex> ProcessHub.is_partitioned?(:my_hub)\n    false","ref":"ProcessHub.html#is_partitioned?/1-example"},{"type":"function","title":"ProcessHub.nodes/2","doc":"Returns a list of nodes where the `ProcessHub` with the given `t:hub_id/0` is running.\n\nNodes where the `ProcessHub` is running with the same `t:hub_id/0` are considered\nto be part of the same cluster.","ref":"ProcessHub.html#nodes/2"},{"type":"function","title":"Example - ProcessHub.nodes/2","doc":"iex> ProcessHub.nodes(:my_hub, [:include_local])\n    [:remote_node]","ref":"ProcessHub.html#nodes/2-example"},{"type":"function","title":"ProcessHub.process_registry/1","doc":"Returns all information registered regarding the child processes.\n\nThis function queries results from the local `ets` table and does not make any network calls.","ref":"ProcessHub.html#process_registry/1"},{"type":"function","title":"ProcessHub.start_child/3","doc":"Starts a child process that will be distributed across the cluster.\nThe `t:child_spec()` `:id` must be unique.","ref":"ProcessHub.html#start_child/3"},{"type":"function","title":"Example - ProcessHub.start_child/3","doc":"iex> child_spec = %{id: :my_child, start: {MyProcess, :start_link, []}}\n    iex> ProcessHub.start_child(:my_hub, child_spec)\n    {:ok, :start_initiated}\n\nBy default, the `start_child/3` function is **asynchronous** and returns immediately.\nTo wait for the child to start, you can pass `async_wait: true` to the `opts` argument.\nWhen `async_wait: true`, you **must await** the response from the function.\n\nSee `t:init_opts/0` for more options.","ref":"ProcessHub.html#start_child/3-example"},{"type":"function","title":"Example with synchronous wait - ProcessHub.start_child/3","doc":"The synchronous response includes the status code `:ok` or `:error`, a tuple containing the `t:child_id/0` and\na list of tuples where the first key is the node where the child is started, and the second key is the\n`pid()` of the started child. By default, the list should contain only one tuple, but if the\nredundancy strategy is configured for replicas, it may contain more than one tuple.\n\n    iex> child_spec = %{id: :my_child, start: {MyProcess, :start_link, []}}\n    iex> ProcessHub.start_child(:my_hub, child_spec, [async_wait: true]) |> ProcessHub.await()\n    {:ok, {:my_child, [{:mynode, #PID }]}}","ref":"ProcessHub.html#start_child/3-example-with-synchronous-wait"},{"type":"function","title":"ProcessHub.start_children/3","doc":"Starts multiple child processes that will be distributed across the cluster.\n\nSame as `start_child/3`, except it starts multiple children at once and is more\nefficient than calling `start_child/3` multiple times.\n\n> ##","ref":"ProcessHub.html#start_children/3"},{"type":"function","title":"Warning {: .warning} - ProcessHub.start_children/3","doc":">\n> Using `start_children/3` with `async_wait: true` can lead to timeout errors,\n> especially when the number of children is large.","ref":"ProcessHub.html#start_children/3-warning-warning"},{"type":"function","title":"ProcessHub.start_link/1","doc":"Starts the `ProcessHub` with the given `t:hub_id/0` and settings.\n\nIt is recommended to start the `ProcessHub` under a supervision tree.","ref":"ProcessHub.html#start_link/1"},{"type":"function","title":"ProcessHub.stop/1","doc":"Stops the `ProcessHub` with the given `t:hub_id/0`.","ref":"ProcessHub.html#stop/1"},{"type":"function","title":"ProcessHub.stop_child/3","doc":"Stops a child process in the cluster.\n\nBy default, this function is **asynchronous** and returns immediately.\nYou can wait for the child to stop by passing `async_wait: true` in the `opts` argument.\nWhen `async_wait: true`, you **must await** the response from the function.","ref":"ProcessHub.html#stop_child/3"},{"type":"function","title":"Example - ProcessHub.stop_child/3","doc":"iex> ProcessHub.stop_child(:my_hub, :my_child)\n    {:ok, :stop_initiated}\n\nSee `t:stop_opts/0` for more options.","ref":"ProcessHub.html#stop_child/3-example"},{"type":"function","title":"Example with synchronous wait - ProcessHub.stop_child/3","doc":"iex> ProcessHub.stop_child(:my_hub, :my_child, [async_wait: true]) |> ProcessHub.await()\n    {:ok, {:my_child, [:mynode]}}","ref":"ProcessHub.html#stop_child/3-example-with-synchronous-wait"},{"type":"function","title":"ProcessHub.stop_children/3","doc":"Stops multiple child processes in the cluster.\n\nThis function is similar to `stop_child/3`, but it stops multiple children at once, making it more\nefficient than calling `stop_child/3` multiple times.\n\n> ##","ref":"ProcessHub.html#stop_children/3"},{"type":"function","title":"Warning {: .warning} - ProcessHub.stop_children/3","doc":">\n> Using `stop_children/3` with `async_wait: true` can lead to timeout errors,\n> especially when stopping a large number of child processes.","ref":"ProcessHub.html#stop_children/3-warning-warning"},{"type":"function","title":"ProcessHub.which_children/2","doc":"Works similarly to `Supervisor.which_children/1`, but wraps the result in a tuple\ncontaining the node name and the children.\n\nIt's recommended to use `ProcessHub.process_registry/1` instead when fast lookups\nare required, as it makes no network calls.\n\nAvailable options:\n- `:global` - returns a list of all child processes started by all nodes in the cluster.\n  The return result will be in the format of `[{:node, children}]`.\n- `:local` - returns a list of all child processes started by the local node.\n  The return result will be in the format of `{:node, children}`.","ref":"ProcessHub.html#which_children/2"},{"type":"type","title":"ProcessHub.child_id/0","doc":"The `child_id` defines the name of the child. It is used to identify the child.\nEach child must have a unique `child_id()` in the cluster. A single child may have\nmultiple `pid()`s across the cluster.","ref":"ProcessHub.html#t:child_id/0"},{"type":"type","title":"ProcessHub.child_spec/0","doc":"The `child_spec` defines the specification of a child process.","ref":"ProcessHub.html#t:child_spec/0"},{"type":"type","title":"ProcessHub.hub_id/0","doc":"The `hub_id` defines the name of the hub. It is used to identify the hub.","ref":"ProcessHub.html#t:hub_id/0"},{"type":"type","title":"ProcessHub.init_opts/0","doc":"The `init_opts()` defines the options that can be passed to the `start_children/3`, `start_child/3`,\n`stop_children/3`, and `stop_child/3` functions.\n\n- `:async_wait` - is optional and is used to define whether the function should return another function that\ncan be used to wait for the children to start or stop. The default is `false`.\n- `:timeout` is optional and is used to define the timeout for the function. The timeout option\nshould be used with `async_wait: true`. The default is `5000` (5 seconds).\n- `:check_mailbox` - is optional and is used to define whether the function should clear\nthe mailbox of any existing messages that may overlap. It is recommended to keep this\noption `true` to avoid any unexpected behavior where `start_child/3` or `start_children/3`\ncall timeout but eventually the calling process receives the start responses later. These messages\nwill stay in that process's mailbox, and when the same process calls start child functions again with the\nsame `child_id()`s, it will receive the old responses.This option should be used with `async_wait: true`.\n The default is `true`.\n- `:check_existing` - is optional and is used to define whether the function should check if the children\nare already started. The default is `true`.","ref":"ProcessHub.html#t:init_opts/0"},{"type":"type","title":"ProcessHub.reply_to/0","doc":"The `reply_to` defines the `pid()`s that will receive the response from the hub\nwhen a child is started or stopped.","ref":"ProcessHub.html#t:reply_to/0"},{"type":"type","title":"ProcessHub.stop_opts/0","doc":"The `stop_opts()` defines the options that can be passed to the `stop_children/3` and `stop_child/3` functions.\n\n- `:async_wait` - is optional and is used to define whether the function should return another function that\ncan be used to wait for the children to stop. The default is `false`.\n- `:timeout` is optional and is used to define the timeout for the function. The timeout option\nshould be used with `async_wait: true`. The default is `5000` (5 seconds).","ref":"ProcessHub.html#t:stop_opts/0"},{"type":"type","title":"ProcessHub.t/0","doc":"This is the base configuration structure for the hub and has to be passed to the `start_link/1` function.\n- `:hub_id` is the name of the hub and is required.\n- `:hooks` are optional and are used to define the hooks that can be triggered on specific events.\n- `:redundancy_strategy` is optional and is used to define the strategy for redundancy.\nThe default is `ProcessHub.Strategy.Redundancy.Singularity`.\n- `:migration_strategy` is optional and is used to define the strategy for migration.\nThe default is `ProcessHub.Strategy.Migration.ColdSwap`.\n- `:synchronization_strategy` is optional and is used to define the strategy for synchronization.\nThe default is `ProcessHub.Strategy.Synchronization.PubSub`.\n- `:partition_tolerance_strategy` is optional and is used to define the strategy for partition tolerance.\nThe default is `ProcessHub.Strategy.PartitionTolerance.Divergence`.","ref":"ProcessHub.html#t:t/0"},{"type":"module","title":"ProcessHub.Constant.Event","doc":"Custom events defined as macros.","ref":"ProcessHub.Constant.Event.html"},{"type":"type","title":"ProcessHub.Constant.Event.event_children_registration/0","doc":"Event used when a process has been registered in the ProcessHub registry.","ref":"ProcessHub.Constant.Event.html#t:event_children_registration/0"},{"type":"type","title":"ProcessHub.Constant.Event.event_children_unregistration/0","doc":"Event used when a process has been unregistered from the ProcessHub registry.","ref":"ProcessHub.Constant.Event.html#t:event_children_unregistration/0"},{"type":"type","title":"ProcessHub.Constant.Event.event_cluster_join/0","doc":"Event used when a node joins the ProcessHub cluster.","ref":"ProcessHub.Constant.Event.html#t:event_cluster_join/0"},{"type":"type","title":"ProcessHub.Constant.Event.event_cluster_leave/0","doc":"Event used when a node leaves the ProcessHub cluster.","ref":"ProcessHub.Constant.Event.html#t:event_cluster_leave/0"},{"type":"type","title":"ProcessHub.Constant.Event.event_distribute_children/0","doc":"Event used when redistributing children to other nodes.","ref":"ProcessHub.Constant.Event.html#t:event_distribute_children/0"},{"type":"type","title":"ProcessHub.Constant.Event.event_sync_remote_children/0","doc":"Event indicating that a remote node is trying to sync its processes.","ref":"ProcessHub.Constant.Event.html#t:event_sync_remote_children/0"},{"type":"module","title":"ProcessHub.Constant.Hook","doc":"Available hooks for ProcessHub.","ref":"ProcessHub.Constant.Hook.html"},{"type":"function","title":"ProcessHub.Constant.Hook.child_migrated/0","doc":"Hook triggered when a process is migrated to another node.","ref":"ProcessHub.Constant.Hook.html#child_migrated/0"},{"type":"function","title":"ProcessHub.Constant.Hook.cluster_join/0","doc":"Hook triggered when a new node is registered under the ProcessHub cluster.","ref":"ProcessHub.Constant.Hook.html#cluster_join/0"},{"type":"function","title":"ProcessHub.Constant.Hook.cluster_leave/0","doc":"Hook triggered when a node is unregistered from the ProcessHub cluster.","ref":"ProcessHub.Constant.Hook.html#cluster_leave/0"},{"type":"function","title":"ProcessHub.Constant.Hook.post_nodes_redistribution/0","doc":"Hook triggered after processes are redistributed.","ref":"ProcessHub.Constant.Hook.html#post_nodes_redistribution/0"},{"type":"function","title":"ProcessHub.Constant.Hook.pre_nodes_redistribution/0","doc":"Hook triggered before processes are redistributed.","ref":"ProcessHub.Constant.Hook.html#pre_nodes_redistribution/0"},{"type":"function","title":"ProcessHub.Constant.Hook.priority_state_updated/0","doc":"Hook triggered when the priority level of the local event queue has been updated.","ref":"ProcessHub.Constant.Hook.html#priority_state_updated/0"},{"type":"function","title":"ProcessHub.Constant.Hook.registry_pid_inserted/0","doc":"Hook triggered when a new process is registered in the ProcessHub registry.","ref":"ProcessHub.Constant.Hook.html#registry_pid_inserted/0"},{"type":"function","title":"ProcessHub.Constant.Hook.registry_pid_removed/0","doc":"Hook triggered when a process is unregistered from the ProcessHub registry.","ref":"ProcessHub.Constant.Hook.html#registry_pid_removed/0"},{"type":"module","title":"ProcessHub.Constant.PriorityLevel","doc":"Priority levels for the local event queue.","ref":"ProcessHub.Constant.PriorityLevel.html"},{"type":"function","title":"ProcessHub.Constant.PriorityLevel.locked/0","doc":"Priority level for the local event queue when it is locked.","ref":"ProcessHub.Constant.PriorityLevel.html#locked/0"},{"type":"function","title":"ProcessHub.Constant.PriorityLevel.unlocked/0","doc":"Priority level for the local event queue when it is unlocked.","ref":"ProcessHub.Constant.PriorityLevel.html#unlocked/0"},{"type":"type","title":"ProcessHub.Constant.PriorityLevel.priority_level/0","doc":"","ref":"ProcessHub.Constant.PriorityLevel.html#t:priority_level/0"},{"type":"module","title":"ProcessHub.Coordinator","doc":"The `ProcessHub` coordinator module is responsible for coordinating most of the `ProcessHub` events and work.\n\nIn most cases, the coordinator module delegates the work to other service-based\nmodules or handler processes that are created on demand.\n\nEach `ProcessHub` instance has its own coordinator process that handles the\ncoordination. These processes are supervised by the `ProcessHub.Initializer`\nsupervisor.\n\nThe coordinator stores state about the `ProcessHub` instance, such as the cluster nodes.\n\nAdditionally, the coordinator takes care of any periodic tasks required by the\n`ProcessHub` instance, such as `:ets` table cleanups, initial synchronization, propagation, etc.","ref":"ProcessHub.Coordinator.html"},{"type":"function","title":"ProcessHub.Coordinator.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"ProcessHub.Coordinator.html#child_spec/1"},{"type":"function","title":"ProcessHub.Coordinator.handle_continue/2","doc":"","ref":"ProcessHub.Coordinator.html#handle_continue/2"},{"type":"function","title":"ProcessHub.Coordinator.init/1","doc":"","ref":"ProcessHub.Coordinator.html#init/1"},{"type":"function","title":"ProcessHub.Coordinator.start_link/1","doc":"","ref":"ProcessHub.Coordinator.html#start_link/1"},{"type":"type","title":"ProcessHub.Coordinator.t/0","doc":"","ref":"ProcessHub.Coordinator.html#t:t/0"},{"type":"module","title":"ProcessHub.DistributedSupervisor","doc":"The `ProcessHub` distributed supervisor module is responsible for starting and stopping\nthe child processes distributed across the cluster.\n\nEach `ProcessHub` instance has its own distributed supervisor that manages local\nchild processes.","ref":"ProcessHub.DistributedSupervisor.html"},{"type":"function","title":"ProcessHub.DistributedSupervisor.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"ProcessHub.DistributedSupervisor.html#child_spec/1"},{"type":"function","title":"ProcessHub.DistributedSupervisor.has_child?/2","doc":"Returns `true` if the child process is running on local node.","ref":"ProcessHub.DistributedSupervisor.html#has_child?/2"},{"type":"function","title":"ProcessHub.DistributedSupervisor.local_child_ids/1","doc":"Returns the child process ids that are running on local node.","ref":"ProcessHub.DistributedSupervisor.html#local_child_ids/1"},{"type":"function","title":"ProcessHub.DistributedSupervisor.local_pid/2","doc":"Returns the child process pid if it is running on local node.","ref":"ProcessHub.DistributedSupervisor.html#local_pid/2"},{"type":"function","title":"ProcessHub.DistributedSupervisor.start_child/2","doc":"Starts a child process on local node.","ref":"ProcessHub.DistributedSupervisor.html#start_child/2"},{"type":"function","title":"ProcessHub.DistributedSupervisor.start_link/1","doc":"","ref":"ProcessHub.DistributedSupervisor.html#start_link/1"},{"type":"function","title":"ProcessHub.DistributedSupervisor.terminate_child/2","doc":"Stops a child process on local node by first terminating the process and then\ndeleting it from the supervisor child spec list.","ref":"ProcessHub.DistributedSupervisor.html#terminate_child/2"},{"type":"type","title":"ProcessHub.DistributedSupervisor.pname/0","doc":"","ref":"ProcessHub.DistributedSupervisor.html#t:pname/0"},{"type":"module","title":"ProcessHub.Handler.ChildrenAdd.StartHandle","doc":"Handler for starting child processes.","ref":"ProcessHub.Handler.ChildrenAdd.StartHandle.html"},{"type":"function","title":"ProcessHub.Handler.ChildrenAdd.StartHandle.handle/1","doc":"","ref":"ProcessHub.Handler.ChildrenAdd.StartHandle.html#handle/1"},{"type":"type","title":"ProcessHub.Handler.ChildrenAdd.StartHandle.t/0","doc":"","ref":"ProcessHub.Handler.ChildrenAdd.StartHandle.html#t:t/0"},{"type":"module","title":"ProcessHub.Handler.ChildrenAdd.SyncHandle","doc":"Handler for synchronizing added child processes.","ref":"ProcessHub.Handler.ChildrenAdd.SyncHandle.html"},{"type":"function","title":"ProcessHub.Handler.ChildrenAdd.SyncHandle.handle/1","doc":"","ref":"ProcessHub.Handler.ChildrenAdd.SyncHandle.html#handle/1"},{"type":"type","title":"ProcessHub.Handler.ChildrenAdd.SyncHandle.process_startup_result/0","doc":"","ref":"ProcessHub.Handler.ChildrenAdd.SyncHandle.html#t:process_startup_result/0"},{"type":"type","title":"ProcessHub.Handler.ChildrenAdd.SyncHandle.t/0","doc":"","ref":"ProcessHub.Handler.ChildrenAdd.SyncHandle.html#t:t/0"},{"type":"module","title":"ProcessHub.Handler.ChildrenRem.StopHandle","doc":"Handler for stopping child processes.","ref":"ProcessHub.Handler.ChildrenRem.StopHandle.html"},{"type":"function","title":"ProcessHub.Handler.ChildrenRem.StopHandle.handle/1","doc":"","ref":"ProcessHub.Handler.ChildrenRem.StopHandle.html#handle/1"},{"type":"type","title":"ProcessHub.Handler.ChildrenRem.StopHandle.t/0","doc":"","ref":"ProcessHub.Handler.ChildrenRem.StopHandle.html#t:t/0"},{"type":"module","title":"ProcessHub.Handler.ChildrenRem.SyncHandle","doc":"Handler for synchronizing stopped child processes.","ref":"ProcessHub.Handler.ChildrenRem.SyncHandle.html"},{"type":"function","title":"ProcessHub.Handler.ChildrenRem.SyncHandle.handle/1","doc":"","ref":"ProcessHub.Handler.ChildrenRem.SyncHandle.html#handle/1"},{"type":"type","title":"ProcessHub.Handler.ChildrenRem.SyncHandle.t/0","doc":"","ref":"ProcessHub.Handler.ChildrenRem.SyncHandle.html#t:t/0"},{"type":"module","title":"ProcessHub.Handler.ClusterUpdate.NodeDown","doc":"Handler for the node down event.","ref":"ProcessHub.Handler.ClusterUpdate.NodeDown.html"},{"type":"function","title":"ProcessHub.Handler.ClusterUpdate.NodeDown.handle/1","doc":"","ref":"ProcessHub.Handler.ClusterUpdate.NodeDown.html#handle/1"},{"type":"type","title":"ProcessHub.Handler.ClusterUpdate.NodeDown.t/0","doc":"","ref":"ProcessHub.Handler.ClusterUpdate.NodeDown.html#t:t/0"},{"type":"module","title":"ProcessHub.Handler.ClusterUpdate.NodeUp","doc":"Handler for the node up event.","ref":"ProcessHub.Handler.ClusterUpdate.NodeUp.html"},{"type":"function","title":"ProcessHub.Handler.ClusterUpdate.NodeUp.handle/1","doc":"","ref":"ProcessHub.Handler.ClusterUpdate.NodeUp.html#handle/1"},{"type":"type","title":"ProcessHub.Handler.ClusterUpdate.NodeUp.t/0","doc":"","ref":"ProcessHub.Handler.ClusterUpdate.NodeUp.html#t:t/0"},{"type":"module","title":"ProcessHub.Handler.Synchronization.IntervalSyncHandle","doc":"Handler for periodic synchronization.","ref":"ProcessHub.Handler.Synchronization.IntervalSyncHandle.html"},{"type":"function","title":"ProcessHub.Handler.Synchronization.IntervalSyncHandle.handle/1","doc":"","ref":"ProcessHub.Handler.Synchronization.IntervalSyncHandle.html#handle/1"},{"type":"type","title":"ProcessHub.Handler.Synchronization.IntervalSyncHandle.t/0","doc":"","ref":"ProcessHub.Handler.Synchronization.IntervalSyncHandle.html#t:t/0"},{"type":"module","title":"ProcessHub.Handler.Synchronization.IntervalSyncInit","doc":"Handler for initializing synchronization.","ref":"ProcessHub.Handler.Synchronization.IntervalSyncInit.html"},{"type":"function","title":"ProcessHub.Handler.Synchronization.IntervalSyncInit.handle/1","doc":"","ref":"ProcessHub.Handler.Synchronization.IntervalSyncInit.html#handle/1"},{"type":"type","title":"ProcessHub.Handler.Synchronization.IntervalSyncInit.t/0","doc":"","ref":"ProcessHub.Handler.Synchronization.IntervalSyncInit.html#t:t/0"},{"type":"module","title":"ProcessHub.Handler.Synchronization.ProcessEmitHandle","doc":"Handler for emitting process registry data.","ref":"ProcessHub.Handler.Synchronization.ProcessEmitHandle.html"},{"type":"function","title":"ProcessHub.Handler.Synchronization.ProcessEmitHandle.handle/1","doc":"","ref":"ProcessHub.Handler.Synchronization.ProcessEmitHandle.html#handle/1"},{"type":"type","title":"ProcessHub.Handler.Synchronization.ProcessEmitHandle.t/0","doc":"","ref":"ProcessHub.Handler.Synchronization.ProcessEmitHandle.html#t:t/0"},{"type":"module","title":"ProcessHub.Initializer","doc":"The main `ProcessHub` initializer supervisor module.\n\nThis module is responsible for starting all the child processes of `ProcessHub`.","ref":"ProcessHub.Initializer.html"},{"type":"function","title":"ProcessHub.Initializer.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"ProcessHub.Initializer.html#child_spec/1"},{"type":"function","title":"ProcessHub.Initializer.start_link/1","doc":"Starts a `ProcessHub` instance with all its children.","ref":"ProcessHub.Initializer.html#start_link/1"},{"type":"function","title":"ProcessHub.Initializer.stop/1","doc":"Starts a `ProcessHub` instance with all its children.","ref":"ProcessHub.Initializer.html#stop/1"},{"type":"module","title":"ProcessHub.Service.Cluster","doc":"`ProcessHub` instances with the same `hub_id` will automatically form a cluster.\nThe cluster service provides API functions for managing the cluster.","ref":"ProcessHub.Service.Cluster.html"},{"type":"function","title":"ProcessHub.Service.Cluster.add_cluster_node/2","doc":"Adds a new node to the cluster and returns the new list of nodes.","ref":"ProcessHub.Service.Cluster.html#add_cluster_node/2"},{"type":"function","title":"ProcessHub.Service.Cluster.new_node?/2","doc":"Returns a boolean indicating whether the node exists in the cluster.","ref":"ProcessHub.Service.Cluster.html#new_node?/2"},{"type":"function","title":"ProcessHub.Service.Cluster.nodes/2","doc":"Returns a list of nodes in the cluster.","ref":"ProcessHub.Service.Cluster.html#nodes/2"},{"type":"function","title":"ProcessHub.Service.Cluster.propagate_self/2","doc":"Sends a cluster join event to the remote node.","ref":"ProcessHub.Service.Cluster.html#propagate_self/2"},{"type":"function","title":"ProcessHub.Service.Cluster.rem_cluster_node/2","doc":"Removes a node from the cluster and returns the new list of nodes.","ref":"ProcessHub.Service.Cluster.html#rem_cluster_node/2"},{"type":"module","title":"ProcessHub.Service.Dispatcher","doc":"The dispatcher service provides API functions for dispatching events.","ref":"ProcessHub.Service.Dispatcher.html"},{"type":"function","title":"ProcessHub.Service.Dispatcher.children_start/2","doc":"Sends the coordinator process a message to start the child processes passed in.","ref":"ProcessHub.Service.Dispatcher.html#children_start/2"},{"type":"function","title":"ProcessHub.Service.Dispatcher.children_stop/2","doc":"Sends the coordinator process a message to stop the child processes passed in.","ref":"ProcessHub.Service.Dispatcher.html#children_stop/2"},{"type":"function","title":"ProcessHub.Service.Dispatcher.propagate_event/5","doc":"Propagates the event to the local or global event queue.","ref":"ProcessHub.Service.Dispatcher.html#propagate_event/5"},{"type":"function","title":"ProcessHub.Service.Dispatcher.reply_respondents/5","doc":"Sends a message to all the respondents who are waiting for a response.","ref":"ProcessHub.Service.Dispatcher.html#reply_respondents/5"},{"type":"module","title":"ProcessHub.Service.Distributor","doc":"The distributor service provides API functions for distributing child processes.","ref":"ProcessHub.Service.Distributor.html"},{"type":"function","title":"ProcessHub.Service.Distributor.child_redist_init/4","doc":"Initiates process redistribution.","ref":"ProcessHub.Service.Distributor.html#child_redist_init/4"},{"type":"function","title":"ProcessHub.Service.Distributor.child_terminate/3","doc":"Terminates a child process and propagates the termination to rest of the cluster.","ref":"ProcessHub.Service.Distributor.html#child_terminate/3"},{"type":"function","title":"ProcessHub.Service.Distributor.init_children/3","doc":"Initiates processes startup.","ref":"ProcessHub.Service.Distributor.html#init_children/3"},{"type":"function","title":"ProcessHub.Service.Distributor.stop_children/3","doc":"Initiates processes shutdown.","ref":"ProcessHub.Service.Distributor.html#stop_children/3"},{"type":"function","title":"ProcessHub.Service.Distributor.which_children_global/2","doc":"Return a list of all child processes started by all nodes in the cluster.","ref":"ProcessHub.Service.Distributor.html#which_children_global/2"},{"type":"function","title":"ProcessHub.Service.Distributor.which_children_local/2","doc":"Returns all child processes started by the local node.\n\nWorks similar to `Supervisor.which_children/1` but returns a list of tuple\nwhere the first element is the node name and the second child processes started under the node.","ref":"ProcessHub.Service.Distributor.html#which_children_local/2"},{"type":"module","title":"ProcessHub.Service.HookManager","doc":"The hook manager service provides API functions for managing hook dispatching,\nregistration, and lookup.","ref":"ProcessHub.Service.HookManager.html"},{"type":"function","title":"ProcessHub.Service.HookManager.cache_key/0","doc":"Returns the cache key for the hook manager.","ref":"ProcessHub.Service.HookManager.html#cache_key/0"},{"type":"function","title":"ProcessHub.Service.HookManager.dispatch_hook/3","doc":"Dispatches the hook to the registered handlers and passes the hook data as an argument.\n\nIt is possible to register a hook handler with a wildcard argument `:_` which\nwill be replaced with the hook data when the hook is dispatched.","ref":"ProcessHub.Service.HookManager.html#dispatch_hook/3"},{"type":"function","title":"ProcessHub.Service.HookManager.dispatch_hooks/2","doc":"Dispatches multiple hooks to the registered handlers.","ref":"ProcessHub.Service.HookManager.html#dispatch_hooks/2"},{"type":"function","title":"ProcessHub.Service.HookManager.register_handler/3","doc":"Registers a new hook handler.","ref":"ProcessHub.Service.HookManager.html#register_handler/3"},{"type":"function","title":"ProcessHub.Service.HookManager.registered_handlers/1","doc":"Returns all registered hook handlers sorted by hook key.","ref":"ProcessHub.Service.HookManager.html#registered_handlers/1"},{"type":"type","title":"ProcessHub.Service.HookManager.hook/0","doc":"","ref":"ProcessHub.Service.HookManager.html#t:hook/0"},{"type":"type","title":"ProcessHub.Service.HookManager.hook_key/0","doc":"","ref":"ProcessHub.Service.HookManager.html#t:hook_key/0"},{"type":"type","title":"ProcessHub.Service.HookManager.hooks/0","doc":"","ref":"ProcessHub.Service.HookManager.html#t:hooks/0"},{"type":"module","title":"ProcessHub.Service.LocalStorage","doc":"The local storage service provides API functions for managing local storage.\nThis is mainly used for storing data that is not required to be persisted.\n\nLocal storage is implemented using ETS tables and is cleared periodically.","ref":"ProcessHub.Service.LocalStorage.html"},{"type":"function","title":"ProcessHub.Service.LocalStorage.exists?/2","doc":"Returns a boolean indicating whether the key exists in local storage.","ref":"ProcessHub.Service.LocalStorage.html#exists?/2"},{"type":"function","title":"ProcessHub.Service.LocalStorage.get/2","doc":"Returns the value for the key in local storage.\n\nIf the key does not exist, nil is returned; otherwise, the value is returned in\nthe format of a tuple: `{key, value, ttl}`.","ref":"ProcessHub.Service.LocalStorage.html#get/2"},{"type":"function","title":"ProcessHub.Service.LocalStorage.insert/4","doc":"Inserts the key and value into local storage.\n\nIt is possible to set a time to live (TTL) for the key. If the TTL is set to\nnil, the key will not expire.\n\nThe `ttl` value is expected to be a positive integer in milliseconds or `nil`.","ref":"ProcessHub.Service.LocalStorage.html#insert/4"},{"type":"module","title":"ProcessHub.Service.Mailbox","doc":"The messenger service provides API functions for receiving messages from other processes.","ref":"ProcessHub.Service.Mailbox.html"},{"type":"function","title":"ProcessHub.Service.Mailbox.receive_child_resp/5","doc":"Waits for multiple child response messages.","ref":"ProcessHub.Service.Mailbox.html#receive_child_resp/5"},{"type":"function","title":"ProcessHub.Service.Mailbox.receive_response/6","doc":"Receives a single child response message.","ref":"ProcessHub.Service.Mailbox.html#receive_response/6"},{"type":"function","title":"ProcessHub.Service.Mailbox.receive_start_resp/2","doc":"Waits for multiple child process startup results.","ref":"ProcessHub.Service.Mailbox.html#receive_start_resp/2"},{"type":"function","title":"ProcessHub.Service.Mailbox.receive_stop_resp/2","doc":"Waits for multiple child process termination results.","ref":"ProcessHub.Service.Mailbox.html#receive_stop_resp/2"},{"type":"module","title":"ProcessHub.Service.ProcessRegistry","doc":"The process registry service provides API functions for managing the process registry.","ref":"ProcessHub.Service.ProcessRegistry.html"},{"type":"function","title":"ProcessHub.Service.ProcessRegistry.bulk_delete/2","doc":"Deletes information about multiple child processes from the registry.\n\nCalling this function will dispatch the `:registry_pid_remove_hook` hooks unless\nthe `:skip_hooks` option is set to `true`.","ref":"ProcessHub.Service.ProcessRegistry.html#bulk_delete/2"},{"type":"function","title":"ProcessHub.Service.ProcessRegistry.bulk_insert/2","doc":"Inserts information about multiple child processes into the registry.\n\nCalling this function will dispatch the `:registry_pid_insert_hook` hook unless the `:skip_hooks` option is set to `true`.","ref":"ProcessHub.Service.ProcessRegistry.html#bulk_insert/2"},{"type":"function","title":"ProcessHub.Service.ProcessRegistry.clear_all/1","doc":"Deletes all objects from the process registry.","ref":"ProcessHub.Service.ProcessRegistry.html#clear_all/1"},{"type":"function","title":"ProcessHub.Service.ProcessRegistry.contains_children/2","doc":"Returns a list of child_ids that match the given `child_ids` variable.","ref":"ProcessHub.Service.ProcessRegistry.html#contains_children/2"},{"type":"function","title":"ProcessHub.Service.ProcessRegistry.delete/3","doc":"Deletes information about a child process from the registry.\n\nCalling this function will dispatch the `:registry_pid_remove_hook` hook unless the `:skip_hooks` option is set to `true`.","ref":"ProcessHub.Service.ProcessRegistry.html#delete/3"},{"type":"function","title":"ProcessHub.Service.ProcessRegistry.insert/4","doc":"Inserts information about a child process into the registry.\n\nCalling this function will dispatch the `:registry_pid_insert_hook` hook unless the `:skip_hooks` option is set to `true`.","ref":"ProcessHub.Service.ProcessRegistry.html#insert/4"},{"type":"function","title":"ProcessHub.Service.ProcessRegistry.local_child_specs/1","doc":"Returns a list of child specs registered under the local node.","ref":"ProcessHub.Service.ProcessRegistry.html#local_child_specs/1"},{"type":"function","title":"ProcessHub.Service.ProcessRegistry.local_children/1","doc":"Returns information about processes registered under the local node.","ref":"ProcessHub.Service.ProcessRegistry.html#local_children/1"},{"type":"function","title":"ProcessHub.Service.ProcessRegistry.lookup/2","doc":"Return the child_spec, nodes, and pids for the given child_id.","ref":"ProcessHub.Service.ProcessRegistry.html#lookup/2"},{"type":"function","title":"ProcessHub.Service.ProcessRegistry.registry/1","doc":"Returns information about all registered processes.","ref":"ProcessHub.Service.ProcessRegistry.html#registry/1"},{"type":"function","title":"ProcessHub.Service.ProcessRegistry.registry_name/1","doc":"Returns the process registry table identifier.","ref":"ProcessHub.Service.ProcessRegistry.html#registry_name/1"},{"type":"type","title":"ProcessHub.Service.ProcessRegistry.registry/0","doc":"","ref":"ProcessHub.Service.ProcessRegistry.html#t:registry/0"},{"type":"module","title":"ProcessHub.Service.Ring","doc":"The Ring service provides API functions for managing the hash ring.","ref":"ProcessHub.Service.Ring.html"},{"type":"function","title":"ProcessHub.Service.Ring.add_node/2","doc":"Adds a new node to the passed-in `hash_ring` and returns the new hash ring.","ref":"ProcessHub.Service.Ring.html#add_node/2"},{"type":"function","title":"ProcessHub.Service.Ring.get_ring/1","doc":"Returns the hash ring instance belonging to the given `hub_id`.","ref":"ProcessHub.Service.Ring.html#get_ring/1"},{"type":"function","title":"ProcessHub.Service.Ring.key_to_node/3","doc":"Determines which node the given `child_id` belongs to.","ref":"ProcessHub.Service.Ring.html#key_to_node/3"},{"type":"function","title":"ProcessHub.Service.Ring.key_to_nodes/3","doc":"Determines which nodes the given `child_id` belongs to.\n\nThe `replication_factor` determines how many nodes to return.","ref":"ProcessHub.Service.Ring.html#key_to_nodes/3"},{"type":"function","title":"ProcessHub.Service.Ring.nodes/1","doc":"Returns a list of all nodes in the hash ring.","ref":"ProcessHub.Service.Ring.html#nodes/1"},{"type":"function","title":"ProcessHub.Service.Ring.remove_node/2","doc":"Removes a node from the hash ring and returns the new hash ring.","ref":"ProcessHub.Service.Ring.html#remove_node/2"},{"type":"module","title":"ProcessHub.Service.State","doc":"The state service provides API functions for managing the state of the hub and\nlocking/unlocking the local event handler.","ref":"ProcessHub.Service.State.html"},{"type":"function","title":"ProcessHub.Service.State.is_locked?/1","doc":"Returns a boolean indicating whether the hub is locked.","ref":"ProcessHub.Service.State.html#is_locked?/1"},{"type":"function","title":"ProcessHub.Service.State.is_partitioned?/1","doc":"Returns a boolean indicating whether the hub cluster is partitioned.","ref":"ProcessHub.Service.State.html#is_partitioned?/1"},{"type":"function","title":"ProcessHub.Service.State.lock_local_event_handler/2","doc":"Locks the local event handler and dispatches the priority state updated hook.","ref":"ProcessHub.Service.State.html#lock_local_event_handler/2"},{"type":"function","title":"ProcessHub.Service.State.toggle_quorum_failure/1","doc":"Locks the local event handler and kills the local distributed supervisor.","ref":"ProcessHub.Service.State.html#toggle_quorum_failure/1"},{"type":"function","title":"ProcessHub.Service.State.toggle_quorum_success/1","doc":"Unlocks the local event handler and restarts the local distributed supervisor.","ref":"ProcessHub.Service.State.html#toggle_quorum_success/1"},{"type":"function","title":"ProcessHub.Service.State.unlock_local_event_handler/1","doc":"Unlocks the local event handler and dispatches the priority state updated hook.","ref":"ProcessHub.Service.State.html#unlock_local_event_handler/1"},{"type":"module","title":"ProcessHub.Service.Synchronizer","doc":"The synchronizer service provides API functions for synchronizing process\nregistry data between nodes.","ref":"ProcessHub.Service.Synchronizer.html"},{"type":"function","title":"ProcessHub.Service.Synchronizer.append_data/2","doc":"Appends remote data to the local process registry.","ref":"ProcessHub.Service.Synchronizer.html#append_data/2"},{"type":"function","title":"ProcessHub.Service.Synchronizer.detach_data/2","doc":"Detaches remote data from the local process registry.","ref":"ProcessHub.Service.Synchronizer.html#detach_data/2"},{"type":"function","title":"ProcessHub.Service.Synchronizer.exec_interval_sync/4","doc":"","ref":"ProcessHub.Service.Synchronizer.html#exec_interval_sync/4"},{"type":"function","title":"ProcessHub.Service.Synchronizer.local_sync_data/1","doc":"Returns local node's process registry data used for synchronization.","ref":"ProcessHub.Service.Synchronizer.html#local_sync_data/1"},{"type":"protocol","title":"ProcessHub.Strategy.Migration.Base","doc":"The migration strategy protocol provides API functions for migrating child processes.","ref":"ProcessHub.Strategy.Migration.Base.html"},{"type":"function","title":"ProcessHub.Strategy.Migration.Base.handle_migration/5","doc":"Migrates processes from the local to the remote node.\n\nProcess migration happens when a new node joins the `ProcessHub` cluster, and some of the\nlocal processes are moved to the newly connected node. This also requires the processes\nto be terminated on the local node.\n\nThe redundancy strategy will deal with scenarios where processes are not terminated locally\nand are duplicated on the new node.","ref":"ProcessHub.Strategy.Migration.Base.html#handle_migration/5"},{"type":"type","title":"ProcessHub.Strategy.Migration.Base.t/0","doc":"All the types that implement this protocol.","ref":"ProcessHub.Strategy.Migration.Base.html#t:t/0"},{"type":"module","title":"ProcessHub.Strategy.Migration.ColdSwap","doc":"The cold swap migration strategy implements the `ProcessHub.Strategy.Migration.Base` protocol.\nIt provides a migration strategy where the local process is terminated before starting it on\nthe remote node.\n\nCold swap is a safe strategy if we want to ensure that the child process is not\nrunning on multiple nodes at the same time.\n\nThis is the default strategy for process migration.","ref":"ProcessHub.Strategy.Migration.ColdSwap.html"},{"type":"type","title":"ProcessHub.Strategy.Migration.ColdSwap.t/0","doc":"The cold swap migration struct.\n\nThis struct does not contain any configuration options.","ref":"ProcessHub.Strategy.Migration.ColdSwap.html#t:t/0"},{"type":"module","title":"ProcessHub.Strategy.Migration.HotSwap","doc":"The hot swap migration strategy implements the `ProcessHub.Strategy.Migration.Base` protocol.\nIt provides a migration strategy where the local process is terminated after the new one is\nstarted on the remote node.\n\nIn the following text, we will refer to the process that is being terminated after the migration\nas the peer process.\n\nThe hot swap strategy is useful when we want to ensure that there is no downtime when migrating\nthe child process to the remote node. It also provides a way to ensure that the state of the\nprocess is synchronized before terminating the peer process.\n\nTo pass the process state to the newly started process on the remote node, the\n`:handover` option must be set to `true` in the migration strategy configuration, and\nnecessary messages must be handled in the processes.\n\n> ##","ref":"ProcessHub.Strategy.Migration.HotSwap.html"},{"type":"module","title":"Using the handover option {: .warning} - ProcessHub.Strategy.Migration.HotSwap","doc":">\n> When the `:handover` option is set to `true`, the peer process must handle the following message:\n> `{:process_hub, :handover_start, startup_resp, from}`.\n\nExample migration with handover using `GenServer`:\n\n```elixir\ndef handle_info({:process_hub, :handover_start, startup_resp, from}, state) do\n  case startup_resp do\n    {:ok, pid} ->\n      # Send the state to the remote process.\n      Process.send(pid, {:process_hub, :handover, state}, [])\n\n      # Signal the handler process that the state handover has been handled.\n      Process.send(from, {:process_hub, :retention_handled}, [])\n\n    _ ->\n      nil\n  end\n\n  {:noreply, state}\nend\n\ndef handle_info({:process_hub, :handover, handover_state}, _state) do\n  {:noreply, handover_state}\nend\n```\n\n> ##","ref":"ProcessHub.Strategy.Migration.HotSwap.html#module-using-the-handover-option-warning"},{"type":"module","title":"Use HotSwap macro to provide the handover callbacks automatically.  {: .info} - ProcessHub.Strategy.Migration.HotSwap","doc":"> It's convenient to use the `HotSwap` macro to provide the handover callbacks automatically.\n>\n> ```elixir\n> use ProcessHub.Strategy.Migration.HotSwap\n> ```","ref":"ProcessHub.Strategy.Migration.HotSwap.html#module-use-hotswap-macro-to-provide-the-handover-callbacks-automatically-info"},{"type":"type","title":"ProcessHub.Strategy.Migration.HotSwap.t/0","doc":"Hot-swap migration strategy configuration.\nAvailable options:\n- `:retention` - An integer value in milliseconds or the `:none` atom is used to specify how long\n  the peer process should be kept alive after a new child process has been started on the remote node.\n  This option is used to ensure that the peer process has had enough time to perform any cleanup\n  or state synchronization before the local process is terminated. It is **strongly recommended**\n  to use this option.\n  The default value is `:none`.\n\n- `:handover` - A boolean value. If set to `true`, the processes involved in the migration\n  must handle the following message: `{:process_hub, :handover_start, startup_resp, from}`.\n  This message will be sent to the peer process that will be terminated after the migration.\n  The variable `startup_resp` will contain the response from the `ProcessHub.DistributedSupervisor.start_child/2` function, and if\n  successful, it will be `{:ok, pid}`. The PID can be used to send the state of the process to the\n  remote process.\n  If the `retention` option is used, then the peer process has to signal the handler process\n  that the state has been handled; otherwise, the handler will wait until the retention time has passed.\n  The handler PID is passed in the `from` variable.\n  The default value is `false`.","ref":"ProcessHub.Strategy.Migration.HotSwap.html#t:t/0"},{"type":"protocol","title":"ProcessHub.Strategy.PartitionTolerance.Base","doc":"The partition tolerance strategy protocol defines the behavior for handling\nnode up and down events in the `ProcessHub` cluster.","ref":"ProcessHub.Strategy.PartitionTolerance.Base.html"},{"type":"function","title":"ProcessHub.Strategy.PartitionTolerance.Base.handle_node_down/4","doc":"This function is called when a new node leaves the `ProcessHub` cluster.","ref":"ProcessHub.Strategy.PartitionTolerance.Base.html#handle_node_down/4"},{"type":"function","title":"ProcessHub.Strategy.PartitionTolerance.Base.handle_node_up/4","doc":"This function is called when a new node joins the `ProcessHub` cluster.","ref":"ProcessHub.Strategy.PartitionTolerance.Base.html#handle_node_up/4"},{"type":"function","title":"ProcessHub.Strategy.PartitionTolerance.Base.handle_startup/3","doc":"This function is called when `ProcessHub` is starting up.","ref":"ProcessHub.Strategy.PartitionTolerance.Base.html#handle_startup/3"},{"type":"type","title":"ProcessHub.Strategy.PartitionTolerance.Base.t/0","doc":"All the types that implement this protocol.","ref":"ProcessHub.Strategy.PartitionTolerance.Base.html#t:t/0"},{"type":"module","title":"ProcessHub.Strategy.PartitionTolerance.Divergence","doc":"The divergence strategy for partition tolerance is used when the `ProcessHub` cluster is not\nconcerned with partition failures.\n\nWhen a node leaves or joins the `ProcessHub` cluster, the divergence strategy keeps the `ProcessHub`\nrunning no matter what.\n\n> #### `In case of network partition` {: .info}\n> The divergence strategy does not handle network partitions in any way. When a network\n> partition occurs, the `ProcessHub` cluster will continue to run on both sides of the partition.\n\nThis is the default strategy for partition tolerance.","ref":"ProcessHub.Strategy.PartitionTolerance.Divergence.html"},{"type":"type","title":"ProcessHub.Strategy.PartitionTolerance.Divergence.t/0","doc":"Divergence strategy configuration.","ref":"ProcessHub.Strategy.PartitionTolerance.Divergence.html#t:t/0"},{"type":"module","title":"ProcessHub.Strategy.PartitionTolerance.DynamicQuorum","doc":"The dynamic quorum strategy provides a dynamic way to handle network partitions in the `ProcessHub` cluster.\n\nThis strategy is suitable for clusters that do not know the number of nodes in the cluster beforehand.\n\nThe dynamic quorum strategy uses a cache to store the quorum log. The quorum log is a list of tuples\ncontaining the node name and the timestamp when the connection to the node was lost.\nThis log is used to calculate the quorum size, which is measured in percentage.\nIf the calculated quorum size is less than the `quorum_size` defined in the configuration,\nthe `ProcessHub` will be considered to be in a network partition.\nIn such a case, the `ProcessHub` will terminate its distributed supervisor process,\n`ProcessHub.DistributedSupervisor`, and all its children. Also, the local event queue will be\nlocked by increasing the priority level.\n\nThe quorum size is calculated by the following formula:\n\n    quorum_size = (connected_nodes / (connected_nodes + down_nodes)) * 100\n\n    # The `connected_nodes` are the number of nodes that are currently connected to the `ProcessHub` cluster.\n    # The `down_nodes` are the number of nodes that are listed under the quorum log.\n\nThe `threshold_time` is used to determine how long a node should be listed in the quorum log\nbefore it is removed from the quorum log. This means if we have a `threshold_time` of 10 seconds,\nand we lose one node every 10 seconds until we have lost all nodes, the system won't be\nconsidered to be in a network partition, regardless of whether the lost nodes are lost due to network\npartition or not. If we lose all those nodes in less than 10 seconds, the system will be considered\nto be in a network partition.\n\n> #### `Scaling down` {: .error}\n> When scaling down the cluster, make sure to scale down one node at a time\n> and wait for the `threshold_time` before scaling down the next node.\n>\n> Do not scale down too many nodes at once because the system may think that it is in a network partition\n> and terminate the `ProcessHub.DistributedSupervisor` process and all its children.","ref":"ProcessHub.Strategy.PartitionTolerance.DynamicQuorum.html"},{"type":"type","title":"ProcessHub.Strategy.PartitionTolerance.DynamicQuorum.t/0","doc":"Dynamic quorum strategy configuration.\n\n- `quorum_size` - The quorum size is measured in percentage. This is the required quorum size.\nFor example, `50` means that if we have 10 nodes in the cluster, the system will be\nconsidered to be in a network partition if we lose 6 (60% of the cluster) nodes or more within\nthe `threshold_time` period.\n- `threshold_time` - The threshold time is measured in seconds. This is how long a node should be\nlisted in the quorum log before it is removed from the quorum log.","ref":"ProcessHub.Strategy.PartitionTolerance.DynamicQuorum.html#t:t/0"},{"type":"module","title":"ProcessHub.Strategy.PartitionTolerance.StaticQuorum","doc":"The static quorum strategy for partition tolerance is used when the `ProcessHub` cluster is\nconcerned with partition failures, and the cluster size is known.\n\nWhen a node leaves or joins the `ProcessHub` cluster, the static quorum strategy will check if the\nquorum is present. If the quorum is not present, the `ProcessHub` will be considered in a network\npartition, and the distributed supervisor process of the `ProcessHub` will be terminated along with\nall the child processes.","ref":"ProcessHub.Strategy.PartitionTolerance.StaticQuorum.html"},{"type":"type","title":"ProcessHub.Strategy.PartitionTolerance.StaticQuorum.t/0","doc":"Static quorum strategy configuration.\n\n- `quorum_size` - The quorum size is measured in the number of nodes. For example, `3` means that\n  there should be at least 3 nodes in the cluster for the `ProcessHub` to be considered healthy.\n- `startup_confirm` - If set to `true`, the `ProcessHub` will check if the quorum is present\n  when the `ProcessHub` is starting up. If the quorum is not present, it will be\n  considered as a network partition. The default value is `false`.","ref":"ProcessHub.Strategy.PartitionTolerance.StaticQuorum.html#t:t/0"},{"type":"protocol","title":"ProcessHub.Strategy.Redundancy.Base","doc":"The redundancy protocol relies on the `HashRing` library to distribute processes across\nthe cluster and determine which node should be responsible for a given process by its `child_id` key.\n\nIt is possible to start the same process on multiple nodes in the cluster.","ref":"ProcessHub.Strategy.Redundancy.Base.html"},{"type":"function","title":"ProcessHub.Strategy.Redundancy.Base.belongs_to/3","doc":"Determines the nodes that are responsible for the given `child_id` (process).","ref":"ProcessHub.Strategy.Redundancy.Base.html#belongs_to/3"},{"type":"function","title":"ProcessHub.Strategy.Redundancy.Base.handle_post_start/4","doc":"This function is called when `ProcessHub.DistributedSupervisor` has started a new\nchild process, and the strategy can perform any post-start actions.","ref":"ProcessHub.Strategy.Redundancy.Base.html#handle_post_start/4"},{"type":"function","title":"ProcessHub.Strategy.Redundancy.Base.handle_post_update/4","doc":"This function is called when `ProcessHub.DistributedSupervisor` has started a\nreplica of a child process, and the strategy can perform any post-update actions.","ref":"ProcessHub.Strategy.Redundancy.Base.html#handle_post_update/4"},{"type":"function","title":"ProcessHub.Strategy.Redundancy.Base.replication_factor/2","doc":"Returns the replication factor for the given strategy struct. This is the number of replicas\nthat the process will be started with.","ref":"ProcessHub.Strategy.Redundancy.Base.html#replication_factor/2"},{"type":"type","title":"ProcessHub.Strategy.Redundancy.Base.t/0","doc":"All the types that implement this protocol.","ref":"ProcessHub.Strategy.Redundancy.Base.html#t:t/0"},{"type":"module","title":"ProcessHub.Strategy.Redundancy.Replication","doc":"The replication strategy allows for multiple instances of a process to be started\nacross the cluster. The number of instances is determined by the `replication_factor`\noption.\n\nThe replication strategy also comes with a `replication_model` option, which determines\nwhich instances of the process are active and which are passive.\nImagine a scenario where you have a process that is responsible for handling streams of\ndata. You may want to have multiple instances of this process running across the cluster,\nbut only one of them should be active at any given time and insert data into the database.\nThe passive instances of the process should be ready to take over if the active instance\nfails.\n\nThis strategy also allows replication on all nodes in the cluster. This is done by setting\nthe `replication_factor` option to `:cluster_size`.\n\n`ProcessHub` can notify the process when it is active or passive by sending a `:redundancy_signal`.\n\n> #### `Using the :redundancy_signal option` {: .info}\n> When using the `:redundancy_signal` option, make sure that the processes are handling\n> the message `{:process_hub, :redundancy_signal, mode}`, where the `mode` variable is either\n> `:active` or `:passive`.\n\nExample `GenServer` process handling the `:redundancy_signal`:\n    def handle_info({:process_hub, :redundancy_signal, mode}, state) do\n      # Update the state with the new mode and do something differently.\n      {:noreply, Map.put(state, :replication_mode, mode)}\n    end","ref":"ProcessHub.Strategy.Redundancy.Replication.html"},{"type":"type","title":"ProcessHub.Strategy.Redundancy.Replication.t/0","doc":"Replication strategy options.\n\n- `replication_factor` - The number of instances of a single process across the cluster.\nThis option can be positive integer or `cluster_size` atom to replicate process on all nodes.\nDefault value is `2`.\n- `replication_model` - This option determines the mode in which the instances of processes are started.\nDefault value is `:active_active`.\n  - `:active_active` - All instances of a process across the cluster are equal.\n  - `:active_passive` - Only one instance of a process across the cluster is active; the rest are passive.\n    Remaining replicas are started as passive processes.\n- `redundancy_signal` - This option determines when a process should be notified of its replication mode.\nDefault value is `:none`.\n  - `:none` - No notifications are sent.\n  - `:active_active` - Only active processes are notified.\n  - `:active_passive` - Only passive processes are notified.\n  - `:all` - All processes are notified.","ref":"ProcessHub.Strategy.Redundancy.Replication.html#t:t/0"},{"type":"module","title":"ProcessHub.Strategy.Redundancy.Singularity","doc":"The Singularity strategy starts a single instance of a process and creates no replicas on\nother nodes. This is the default strategy.","ref":"ProcessHub.Strategy.Redundancy.Singularity.html"},{"type":"type","title":"ProcessHub.Strategy.Redundancy.Singularity.t/0","doc":"No options are available.","ref":"ProcessHub.Strategy.Redundancy.Singularity.html#t:t/0"},{"type":"protocol","title":"ProcessHub.Strategy.Synchronization.Base","doc":"This protocol defines the behavior of a synchronization strategy.","ref":"ProcessHub.Strategy.Synchronization.Base.html"},{"type":"function","title":"ProcessHub.Strategy.Synchronization.Base.handle_propagation/4","doc":"This function handles the propagation messages sent by `ProcessHub.Strategy.Synchronization.Base.propagate/5`.\n\nIt saves the process data that was propagated to the local process registry.","ref":"ProcessHub.Strategy.Synchronization.Base.html#handle_propagation/4"},{"type":"function","title":"ProcessHub.Strategy.Synchronization.Base.handle_synchronization/4","doc":"Handles the periodic synchronization of the process registry.","ref":"ProcessHub.Strategy.Synchronization.Base.html#handle_synchronization/4"},{"type":"function","title":"ProcessHub.Strategy.Synchronization.Base.init_sync/3","doc":"Initializes the periodic synchronization of the process registry.","ref":"ProcessHub.Strategy.Synchronization.Base.html#init_sync/3"},{"type":"function","title":"ProcessHub.Strategy.Synchronization.Base.propagate/5","doc":"This function is called when a process has been started on the local node, and the\ninformation about the process is about to be propagated to other nodes.","ref":"ProcessHub.Strategy.Synchronization.Base.html#propagate/5"},{"type":"type","title":"ProcessHub.Strategy.Synchronization.Base.t/0","doc":"All the types that implement this protocol.","ref":"ProcessHub.Strategy.Synchronization.Base.html#t:t/0"},{"type":"module","title":"ProcessHub.Strategy.Synchronization.Gossip","doc":"The Gossip synchronization strategy provides a method for spreading information to other nodes\nwithin the `ProcessHub` cluster. It utilizes a gossip protocol to\nsynchronize the process registry across the cluster.\n\nThe Gossip strategy is most suitable for clusters are large. It scales well but produces\nhigher latency than the PubSub strategy when operating in small clusters.\nWhen the cluster increases in size, Gossip protocol can also save bandwidth compared to PubSub.\n\n> The Gossip strategy works as follows:\n> - The synchronization process is initiated on a single node.\n> - The node collects its own local process registry data and appends it to the synchronization data.\n> - It selects a predefined number of nodes that have not yet added their local registry data.\n> - The node sends the data to the selected nodes.\n> - The nodes append their local registry data to the received data and send it to the next nodes.\n> - When all nodes have added their data to the synchronization data, the message will be sent to\n> nodes that have not yet acknowledged the synchronization ack.\n> - If node receives the synchronization data which contains all nodes data, it will\n> synchronize the data with it's local process registry and forward the data to the next nodes\n> that have not yet acknowledged the synchronization ack.\n> - When all nodes in the cluster have acknowledged the synchronization data, the synchronization\n> process is completed and the reference is invalidated.\n\nEach node also adds a timestamp to the synchronization data. This is used to ensure that\nthe synchronization data is not older than the data that is already in the local process registry.","ref":"ProcessHub.Strategy.Synchronization.Gossip.html"},{"type":"type","title":"ProcessHub.Strategy.Synchronization.Gossip.t/0","doc":"The Gossip strategy configuration options.\n\n* `sync_interval` - The periodic synchronization interval in milliseconds. The default is `15000`.\n* `recipients` - The number of nodes that will receive the synchronization data and propagate it further. The default is `3`.\n* `restricted_init` - If set to `true`, the synchronization process will only be started on a single node.\n  This node is selected by sorting the node names alphabetically and selecting the first node. The default is `true`.","ref":"ProcessHub.Strategy.Synchronization.Gossip.html#t:t/0"},{"type":"module","title":"ProcessHub.Strategy.Synchronization.PubSub","doc":"This PubSub synchronization strategy uses the `:blockade` library to\ndispatch and handle synchronization events. Each `ProcessHub` instance\nhas its own global event queue that is used to dispatch and handle synchronization events.","ref":"ProcessHub.Strategy.Synchronization.PubSub.html"},{"type":"type","title":"ProcessHub.Strategy.Synchronization.PubSub.t/0","doc":"The PubSub synchronization strategy options.\n\n- `sync_interval` - the periodic synchronization interval in milliseconds. Defaults to `15000`.","ref":"ProcessHub.Strategy.Synchronization.PubSub.html#t:t/0"},{"type":"module","title":"ProcessHub.Utility.Bag","doc":"Utility functions.","ref":"ProcessHub.Utility.Bag.html"},{"type":"function","title":"ProcessHub.Utility.Bag.all_messages/0","doc":"Returns all messages in the mailbox.","ref":"ProcessHub.Utility.Bag.html#all_messages/0"},{"type":"function","title":"ProcessHub.Utility.Bag.gen_child_specs/2","doc":"Generates child specs for testing.","ref":"ProcessHub.Utility.Bag.html#gen_child_specs/2"},{"type":"function","title":"ProcessHub.Utility.Bag.hook_erlang_send/3","doc":"Sends hook messages to the given process.","ref":"ProcessHub.Utility.Bag.html#hook_erlang_send/3"},{"type":"function","title":"ProcessHub.Utility.Bag.receive_multiple/3","doc":"Waits and receives multiple messages.","ref":"ProcessHub.Utility.Bag.html#receive_multiple/3"},{"type":"function","title":"ProcessHub.Utility.Bag.timestamp/1","doc":"Returns the current timestamp in the given precision.","ref":"ProcessHub.Utility.Bag.html#timestamp/1"},{"type":"module","title":"ProcessHub.Utility.Name","doc":"Utility functions for generating names processes/ETS tables registered under\nthe `hub_id`.","ref":"ProcessHub.Utility.Name.html"},{"type":"function","title":"ProcessHub.Utility.Name.concat_name/2","doc":"Concatenates the given atoms/binary with the given separator.","ref":"ProcessHub.Utility.Name.html#concat_name/2"},{"type":"function","title":"ProcessHub.Utility.Name.coordinator/1","doc":"The name of main coordinator process.","ref":"ProcessHub.Utility.Name.html#coordinator/1"},{"type":"function","title":"ProcessHub.Utility.Name.distributed_supervisor/1","doc":"The name distributed supervisor process.","ref":"ProcessHub.Utility.Name.html#distributed_supervisor/1"},{"type":"function","title":"ProcessHub.Utility.Name.global_event_queue/1","doc":"The name of the global event queue process.","ref":"ProcessHub.Utility.Name.html#global_event_queue/1"},{"type":"function","title":"ProcessHub.Utility.Name.initializer/1","doc":"The name of the main initializer process.","ref":"ProcessHub.Utility.Name.html#initializer/1"},{"type":"function","title":"ProcessHub.Utility.Name.local_event_queue/1","doc":"The name of the local event queue process.","ref":"ProcessHub.Utility.Name.html#local_event_queue/1"},{"type":"function","title":"ProcessHub.Utility.Name.task_supervisor/1","doc":"The name of the task supervisor process.","ref":"ProcessHub.Utility.Name.html#task_supervisor/1"},{"type":"function","title":"ProcessHub.Utility.Name.worker_queue/1","doc":"","ref":"ProcessHub.Utility.Name.html#worker_queue/1"},{"type":"module","title":"ProcessHub.WorkerQueue","doc":"","ref":"ProcessHub.WorkerQueue.html"},{"type":"function","title":"ProcessHub.WorkerQueue.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"ProcessHub.WorkerQueue.html#child_spec/1"},{"type":"function","title":"ProcessHub.WorkerQueue.handle_call/2","doc":"","ref":"ProcessHub.WorkerQueue.html#handle_call/2"},{"type":"function","title":"ProcessHub.WorkerQueue.init/1","doc":"","ref":"ProcessHub.WorkerQueue.html#init/1"},{"type":"function","title":"ProcessHub.WorkerQueue.start_link/1","doc":"","ref":"ProcessHub.WorkerQueue.html#start_link/1"},{"type":"extras","title":"ProcessHub","doc":"# ProcessHub\n\n![example workflow](https://github.com/alfetahe/process-hub/actions/workflows/elixir.yml/badge.svg)  [![hex.pm version](https://img.shields.io/hexpm/v/coverex.svg?style=flat)](https://hex.pm/packages/process_hub)","ref":"readme.html"},{"type":"extras","title":"Table of Contents - ProcessHub","doc":"* [Description](#description)\n* [Features](#features)\n* [Installation](#installation)\n* [Example usage](#example-usage)\n* [Configurable strategies](#configurable-strategies)\n  * [Redundancy Strategy](#redundancy-strategy)\n  * [Migration Strategy](#migration-strategy)\n  * [Synchronization Strategy](#synchronization-strategy)\n  * [Partition Tolerance Strategy](#partition-tolerance-strategy)\n* [Distribution Strategy](#distribution-strategy)\n* [Cluster Discovery and Formation](#cluster-discovery-and-formation)\n* [Resilience and Reliability](#resilience-and-reliability)\n* [Locking Mechanism](#locking-mechanism)\n* [Hooks](#hooks)\n  * [Available hooks](#available-hooks)\n* [Contributing](#contributing)","ref":"readme.html#table-of-contents"},{"type":"extras","title":"Description - ProcessHub","doc":"Library for building distributed systems that are scalable. It handles the distribution of\nprocesses within a cluster of nodes while providing a globally synchronized process registry.\n\nProcessHub takes care of starting, stopping and monitoring processes in the cluster.\nIt scales automatically when cluster is updated and handles network partitions.\n\nBuilding distributed systems is hard and designing one is all about trade-offs.\nThere are many things to consider and each system has its own requirements. \nThis library aims to be flexible and configurable to suit different use cases.\n\nProcessHub is designed to be **decentralized** in its architecture. It does not rely on a\nsingle node to manage the cluster. Each node in the cluster is considered equal.\nConsensus is achieved by using a hash ring implementation.\n\nDocumentation can be found at [https://hexdocs.pm/process_hub](https://hexdocs.pm/process_hub).\n\n\n> ProcessHub is built with scalability and availability in mind.\n> Most of the operations are asynchronous and non-blocking.\n> It can guarantee **eventual** consistency.\n\nProcessHub provides a set of configurable strategies for building distributed\napplications in Elixir.\n\n> ##","ref":"readme.html#description"},{"type":"extras","title":"ProcessHub requires a distributed node - ProcessHub","doc":"> ProcessHub is distributed in its nature, and for that reason, it needs to\n> **operate in a distributed environment**.\n> This means that the Elixir instance has to be started as a distributed node.\n> For example: `iex --sname mynode --cookie mycookie -S mix`.\n>\n> If the node is not started as a distributed node, starting the `ProcessHub` will fail\n> with the following error: `{:error, :local_node_not_alive}`","ref":"readme.html#processhub-requires-a-distributed-node"},{"type":"extras","title":"Features - ProcessHub","doc":"Main features include:\n- Distributing processes across a cluster of nodes.\n- Distributed and synchronized process registry for fast lookups.\n- Process state handover.\n- Strategies for redundancy handling and process replication.\n- Strategies for handling network failures and partitions automatically.\n- Strategies for handling process migration and synchronization when nodes join/leave\nthe cluster automatically.\n- Hooks for triggering events on specific actions.\n- Automatic hub cluster forming and healing when nodes join or leave the cluster.","ref":"readme.html#features"},{"type":"extras","title":"Installation - ProcessHub","doc":"1. Add `process_hub` to your list of dependencies in `mix.exs`:\n\n    ```elixir\n    def deps do\n      [\n        {:process_hub, \"~> 0.1.3-alpha\"}\n      ]\n    end\n    ```\n\n2. Start the `ProcessHub` supervisor under your application supervision tree:\n\n    ```elixir\n    defmodule MyApp.Application do\n      use Application\n\n      def start(_type, _args) do\n        children = [\n          ProcessHub.child_spec(%ProcessHub{hub_id: :my_hub})\n        ]\n\n        opts = [strategy: :one_for_one, name: MyApp.Supervisor]\n        Supervisor.start_link(children, opts)\n      end\n    end\n    ```\n  It is possible to start multiple hubs under the same supervision tree.\n  Each hub must have a unique `:hub_id`.","ref":"readme.html#installation"},{"type":"extras","title":"Example usage - ProcessHub","doc":"The following example shows how to start 2 elixir nodes, connect them and start processes\n  under the `ProcessHub` cluster. This demonstrates how the processes are distributed within\n  the cluster.\n\n  **Note:** The examples below assume that the `ProcessHub` is already started under the\n  supervision tree. If not please refer to the [Installation](#installation) section.\n\n  **Note:** Make sure you have a GenServer module called `MyProcess` defined in your project.\n  ```elixir\n  defmodule MyProcess do\n    use GenServer\n\n    def start_link() do\n      GenServer.start_link(__MODULE__, nil)\n    end\n\n    def init(_) do\n      {:ok, nil}\n    end\n  end\n  ```\n\n   \n\n  #","ref":"readme.html#example-usage"},{"type":"extras","title":"Node 1 - ProcessHub","doc":"Start the first node with the following command:\n\n  ```bash\n  iex --name node1@127.0.0.1 --cookie mycookie -S mix\n  ```\n\n  ```elixir\n  # Run the following in the iex console to start 5 processes under the hub.\n  iex> ProcessHub.start_children(:my_hub, [\n  ...>  %{id: :some_id1, start: {MyProcess, :start_link, []}},\n  ...>  %{id: :another_id2, start: {MyProcess, :start_link, []}},\n  ...>  %{id: :child_3, start: {MyProcess, :start_link, []}},\n  ...>  %{id: :child_4, start: {MyProcess, :start_link, []}},\n  ...>  %{id: \"the_last_child\", start: {MyProcess, :start_link, []}}\n  ...> ])\n  {:ok, :start_initiated}\n\n  # Check the started processes by running the command below.\n  iex> ProcessHub.which_children(:my_hub, [:global])\n  [\n    \"node1@127.0.0.1\": [\n      {\"the_last_child\", #PID , :worker, [MyProcess]},\n      {:child_4, #PID , :worker, [MyProcess]},\n      {:child_3, #PID , :worker, [MyProcess]},\n      {:another_id2, #PID , :worker, [MyProcess]},\n      {:some_id1, #PID , :worker, [MyProcess]}\n    ]\n  ]\n  ```\n\n  #","ref":"readme.html#node-1"},{"type":"extras","title":"Node 2 - ProcessHub","doc":"We will use this node to connect to the first node and see how the processes are\n  automatically distributed.\n\n  Start the second node.\n  ```bash\n  iex --name node2@127.0.0.1 --cookie mycookie -S mix\n  ```\n\n  ```elixir\n  # Connect the second node to the first node.\n  iex> Node.connect(:\"node1@127.0.0.1\")\n  true\n\n  # Check the started procsses by running the command below and\n  # see how some of the processes are distributed to the second node.\n  iex> ProcessHub.which_children(:my_hub, [:global])\n  [\n    \"node2@127.0.0.1\": [\n      {:child_3, #PID , :worker, [MyProcess]},\n      {:some_id1, #PID , :worker, [MyProcess]}\n    ],\n    \"node1@127.0.0.1\": [\n      {\"the_last_child\", #PID , :worker, [MyProcess]},\n      {:child_4, #PID , :worker, [MyProcess]},\n      {:another_id2, #PID , :worker, [MyProcess]}\n    ]\n  ]\n  ```","ref":"readme.html#node-2"},{"type":"extras","title":"Configurable strategies - ProcessHub","doc":"ProcessHub comes with 9 different strategies that can be used to configure the hub.\nAll strategies are Elixir structs that implement their own base protocol.\n\nIn fact, you can define your own strategies by implementing the base protocols.\n\nWhen configuring the hub, you can pass the strategies as part of the `%ProcessHub{}` struct.\n\nLook at the documentation for each strategy for more information on how to configure them.\n\nAn example can be seen below.\n```elixir\ndefmodule MyApp.Application do\n  use Application\n\n  def start(_type, _args) do\n    children = [process_hub()]\n\n    opts = [strategy: :one_for_one, name: MyApp.Supervisor]\n    Supervisor.start_link(children, opts)\n  end\n\n  defp process_hub() do\n    {ProcessHub, %ProcessHub{\n      hub_id: :my_hub,\n      # Configure the redundancy strategy.\n      redundancy_strategy: %ProcessHub.Strategy.Redundancy.Replication{\n        replication_factor: 2,\n        replication_model: :active_passive,\n        redundancy_signal: :all\n      },\n      # Configure the migration strategy.\n      migration_strategy: %ProcessHub.Strategy.Migration.HotSwap{\n        retention: 2000,\n        handover: true\n      },\n      # Configure the synchronization strategy.\n      synchronization_strategy: %ProcessHub.Strategy.Synchronization.PubSub{\n        sync_interval: 10000\n      },\n      # Configure the partition tolerance strategy.\n      partition_tolerance_strategy: %ProcessHub.Strategy.PartitionTolerance.StaticQuorum{\n        quorum_size: 3\n      }\n    }}\n  end\nend\n```\n\n#","ref":"readme.html#configurable-strategies"},{"type":"extras","title":"Redundancy Strategy - ProcessHub","doc":"`ProcessHub.Strategy.Redundancy.Base` - defines the base protocol for redundancy strategies.\nThis strategy is used to define how many replicas of a process should be started\nacross the cluster. Starting multiple instances of a process across the cluster is useful\nfor redundancy and fault tolerance.\n\nAvailable strategies are:\n- `ProcessHub.Strategy.Redundancy.Singularity` - only 1 process per `child_id` without\nany replicas. This is also the default strategy and contains no special configuration options.\n- `ProcessHub.Strategy.Redundancy.Replication` - starts multiple replicas of a process\nacross the cluster. The number of replicas is defined by the `:replication_factor`\noption. This strategy also supports `:active_active` and `:active_passive` modes.\nMeaning we may have one active process and the rest are passive.\nThe mode is defined by the `:replication_model` option.\nThis information will be passed to the started process.\nThe default mode is `:active_active`, meaning all processes are equal and considered active.\n\n#","ref":"readme.html#redundancy-strategy"},{"type":"extras","title":"Migration Strategy - ProcessHub","doc":"`ProcessHub.Strategy.Migration.Base` - defines the base protocol for migration strategies.\nThis strategy is used to define how the processes are migrated when a node joins or leaves the cluster.\n\nMigration is the process of moving processes from one node to another.\nOne of the reasons why migration happens is when a node leaves the cluster.\nWhen a node leaves the cluster, it is possible that some processes are still running on that\nnode, so these need to be migrated to another node. Also, when a new node joins the\ncluster, other nodes may migrate some processes over to the new node.\n\nAt the moment, there are 2 migration strategies available:\n- `ProcessHub.Strategy.Migration.ColdSwap` - migrate processes by stopping the process\non the old node before starting it on the new node. This is the default strategy and defines no special configuration options.\n- `ProcessHub.Strategy.Migration.HotSwap` - this strategy is used to migrate processes\nby starting the process on the new node before stopping it on the old node.\nThis strategy is useful when we want to avoid any downtime. This strategy is also\nuseful when the process is stateful, and we want to avoid any data loss by handing over\nthe state from the old process to the new process. See the module for handover examples.\n\n#","ref":"readme.html#migration-strategy"},{"type":"extras","title":"Synchronization Strategy - ProcessHub","doc":"`ProcessHub.Strategy.Synchronization.Base` - defines the base protocol for synchronization\nstrategies which define the method that is used to synchronize the process registry.\n\nAvailable strategies are:\n- `ProcessHub.Strategy.Synchronization.PubSub` - uses a publish/subscribe model to synchronize\nthe process registry. Each node in the cluster will subscribe to a topic and publish any changes to the topic.\nThese changes could be events such as adding or removing a process from the registry.\nThis is the default and recommended synchronization strategy for most users.\n- `ProcessHub.Strategy.Synchronization.Gossip` - uses a gossip protocol to synchronize the\nprocess registry. Using this strategy is recommended when the number of nodes in the cluster\nis large. The Gossip strategy selects a predefined number of nodes to gossip with and\nexchange information about the process registry.\nThese selected nodes will choose other nodes to gossip with and so on until all nodes in the\ncluster are synchronized. It has higher latency than the PubSub strategy specially when the cluster\nis rather small.\n\n#","ref":"readme.html#synchronization-strategy"},{"type":"extras","title":"Partition Tolerance Strategy - ProcessHub","doc":"`ProcessHub.Strategy.PartitionTolerance.Base` - defines the base protocol for partition tolerance\nstrategies which define the method that is used to handle network partitions.\n\nAvailable strategies are:\n- `ProcessHub.Strategy.PartitionTolerance.Divergence` - this strategy is used to handle network\npartitions by diverging the cluster into multiple subclusters. Each subcluster will have its\nown hub and will be considered as a separate cluster. This strategy is the default strategy.\nWhen the network partition is healed, the subclusters will merge back into a single cluster.\n- `ProcessHub.Strategy.PartitionTolerance.StaticQuorum` - this strategy is used to handle network\npartitions by using a static quorum. The quorum size is defined by the `:quorum_size` option.\nWhen a partition happens, the `ProcessHub.DistributedSupervisor` process will terminate\nalong with its children. This strategy is useful when the number of nodes in the cluster is\nknown and rather fixed.\n- `ProcessHub.Strategy.PartitionTolerance.DynamicQuorum` - this strategy is used to handle\nnetwork partitions by using a dynamic quorum. The quorum size is defined by the `:quorum_size`\noption and `:threshold_time` option. The system automatically over time adapts to the number of\nnodes in the cluster. When a partition happens, the `ProcessHub.DistributedSupervisor` process\nwill terminate along with its children.\n  > ##","ref":"readme.html#partition-tolerance-strategy"},{"type":"extras","title":"Using DynamicQuorum Strategy - ProcessHub","doc":"> When scaling down too many nodes at once, the system may consider itself to be in a\n  > network partition. Read the documentation for the\n  >`ProcessHub.Strategy.PartitionTolerance.DynamicQuorum` strategy for more information.","ref":"readme.html#using-dynamicquorum-strategy"},{"type":"extras","title":"Distribution Strategy - ProcessHub","doc":"ProcessHub uses consistent hashing to distribute processes. When the cluster is updated, the\nhash ring is recalculated. The recalculation is done in a way that each node is assigned a unique hash value, and they form a **hash ring**. Each node in the cluster keeps track of the ProcessHub cluster and updates its local hash ring accordingly.\n\nTo find the node that the process belongs to, the system will use the hash ring to calculate\nthe hash value of the process ID (`child_id`) and assign it to the node with the closest hash value.\n\nWhen the cluster is updated and the hash ring is recalculated, it does not mean that all\nprocesses will be shuffled. Only the processes that are affected by the change will be redistributed. This is done to avoid unnecessary process migrations.\n\nFor example, when a node leaves the cluster, only the processes that were running on that node\nwill be redistributed. The rest of the processes will stay on the same node. When a new node\njoins the cluster, only some of the processes will be redistributed to the new node, and the rest will stay on the same node.\n\n> The hash ring implementation **does not guarantee** that all processes will always be\n> evenly distributed, but it does its best to distribute them as evenly as possible.\n\nThis strategy is used by default and is not configurable at the moment.","ref":"readme.html#distribution-strategy"},{"type":"extras","title":"Cluster Discovery and Formation - ProcessHub","doc":"ProcessHub monitors connecting and disconnecting nodes and forms a cluster automatically\nfrom the connected nodes that share the same `hub_id`. It's not required to start\nthe `ProcessHub` on all nodes in the cluster.","ref":"readme.html#cluster-discovery-and-formation"},{"type":"extras","title":"Resilience and Reliability - ProcessHub","doc":"ProcessHub uses the `Supervisor` behavior and leverages the features that come with it.\nEach hub starts its own `ProcessHub.DistributedSupervisor` process, which is responsible for\nstarting, stopping, and monitoring the processes in its local cluster.\n\nWhen a process dies unexpectedly, the `ProcessHub.DistributedSupervisor` will restart it\nautomatically.\n\nProcessHub also takes care of validating the `child_spec` before starting it and makes sure\nit's started on the right node that the process belongs to.\nIf the process is being started on the wrong node, the initialization request will be forwarded\nto the correct node.","ref":"readme.html#resilience-and-reliability"},{"type":"extras","title":"Locking Mechanism - ProcessHub","doc":"ProcessHub utilizes the `:blockade` library to provide event-driven communication\nand a locking mechanism.\nIt locks the local event queue by increasing its priority for some operations.\nThis allows the system to queue events and process them in order to preserve data integrity.\nOther events can be processed once the priority level is set back to default.\n\nTo avoid deadlocks, the system places a timeout on the event queue priority and\nrestores it to its original value if the timeout is reached.","ref":"readme.html#locking-mechanism"},{"type":"extras","title":"Hooks - ProcessHub","doc":"Hooks are used to trigger events on specific actions. Hooks can be registered by passing the\nhandlers to the `:hooks` option of the `t:t/0` configuration struct or by inserting them\ndynamically using the `ProcessHub.Service.HookManager` module.\n\nProcessHub heavily uses hooks internally in the integration tests.\n\nHooks have to be in the format of an `mfa` tuple. Basically, they are functions that will be called\nwhen the hook is triggered.\n\nIt is possible to register a hook handler with a wildcard argument `:_`, which will be replaced with the hook data when the hook is dispatched.\n\nExample hook registration using the `t:t/0` configuration struct:\n```elixir\n# Register a hook handler for the `:cluster_join` event with a wildcard argument.\ndefmodule MyApp.Application do\n  use Application\n\n  def start(_type, _args) do\n    children = [\n      ProcessHub.child_spec(%ProcessHub{\n        hub_id: :my_hub,\n        hooks: [\n            ProcessHub.Constant.Hook.cluster_join(), {MyModule, :my_function, [:some_data, :_]}\n        ]\n      })\n    ]\n\n    opts = [strategy: :one_for_one, name: MyApp.Supervisor]\n    Supervisor.start_link(children, opts)\n  end\nend\n```\n\nExample hook registration using the `ProcessHub.Service.HookManager` module:\n```elixir\n# Register a hook handler for the `:cluster_join` event with a wildcard argument.\nProcessHub.Service.HookManager.register_hook(:my_hub, ProcessHub.Constant.Hook.cluster_join(), {MyModule, :my_function, [:some_data, :_]})\n```\n\nExample hook handler:\n```elixir\n# The hook handler should be in the following format:\ndefmodule MyModule do\n  def my_function(:some_data, dynamic_hook_data) do\n    # Do something with the data.\n  end\nend\n```\n\n#","ref":"readme.html#hooks"},{"type":"extras","title":"Available hooks - ProcessHub","doc":"| Event Key                   | Trigger                    | Data                                |\n| ------------                | -------------              | ---------------                     |\n| `cluster_join`              | Node joins the cluster     | `node()`                            |\n| `cluster_leave`             | Node leaves the cluster    | `node()`                            |\n| `registry_pid_inserted`     | Process registered         | `{child_id(), [{node(), pid()}]}` |\n| `registry_pid_removed`      | Process unregistered       | `child_id()`                        |\n| `child_migrated`            | Process migrated           | `{child_id(), node()}`              |\n| `priority_state_updated`    | Priority state updated     | `{priority_level(), list()}`        |\n| `pre_nodes_redistribution`  | Nodes redistribution start | `{:nodeup or :nodedown, node()}`    |\n| `post_nodes_redistribution` | Nodes redistribution end   | `{:nodeup or :nodedown, node()}`    |\n\nSee `ProcessHub.Constant.Hook` module for more information.","ref":"readme.html#available-hooks"},{"type":"extras","title":"Contributing - ProcessHub","doc":"Contributions are welcome and appreciated. If you have any ideas, suggestions, or bugs to report,\nplease open an issue or a pull request on GitHub.","ref":"readme.html#contributing"}],"content_type":"text/markdown"}